##########################################################################
# EXAMPLE R SCRIPT FOR SINGLE-SITE RAINFALL MODEL FITTING
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# This script is intended to illustrate the use of the single-site
# rainfall model fitting routines in file momfit.r. It uses those
# routines to fit a random-parameter Bartlett Lewis model, with
# exponential cell depth distribution, using summary statistics
# for each month of the year at Elmdon. 
#
# Before running this script, make sure you have files 
# elmstats.dat and momfit.r in your current working directory.
#
# The structure of the script is as follows:
#
# 1)	Load the fitting routines. Read the data from file, 
#	and initialise variables relating to fitting properties etc.
# 2)	For each month, fit the model; store the parameter
#	estimates and approximate standard errors.
# 3)	For each parameter, plot the seasonal variation 
#	along with approximate confidence intervals
#
# To run the script, type
#
#	source("fit_demo.r")
#
# at the R prompt. Alternatively, paste the commands into the R window
# one at a time, to see what each one does.
#
# Script author: Richard Chandler (richard@stats.ucl.ac.uk)
# Date:	November 2003 
# Revised: August 2004, January 2006, June 2010, September 2016
##########################################################################
# 1.	Load the fitting routines, read the data and set 
#	everything up. We need to
#
# a)    Change the "Pdry" values in the data to "Pwet",
#       which is the statistic used by the software
#	b)	  Define the fitting properties and their 
#	      associated weights
#	c)    Define an initial estimate of the parameters
#	d)	  Define some bounds on the parameters, to be used 
#       when optimising the objective function
#	e)	  Define some storage for estimates and standard
#		    errors
##########################################################################
source("momfit.r")
elm.data <- read.table("elmstats.dat",header=TRUE)
#
# a) Change Pdry to Pwet
#
elm.data[,9:10] <- 1-elm.data[,9:10]
names(elm.data)[9:10] <- c("Pwet1","Pwet24")
#
# b) The fitting properties we're using are: 1-hour mean, 1-, 6-
#    and 24-hour variance, 1- and 24-hour lag 1 autocorrelation,
#    and 1- and 24-hour proportion of wet intervals. 
#
timescales <- list(Mean=1,Var=c(1,6,24),
                   Ac=matrix(c(1,24,1,1),nrow=2),Ph=c(1,24))
#
# c) Define an initial estimate of the parameters, and give them some 
#    names. The initial guess comes from past experience. The names 
#    will be used to create meaningful plot titles. Note that 
#    the software uses a 2-parameter depth distribution - for
#    an exponential we simply impose the constraint sigma[x] = mu[x]
#    i.e. set parameter 3 equal to 1.
#
theta <- c(0.02,3,1,4,4,0.25,0.1)
names(theta) <- c("lambda","mu[x]","sigma[x]/mu[x]","alpha",
				"alpha/nu","kappa","phi")
#
# d) Now define some lower and upper bounds. These bounds are as 
#    flexible as possible. For the random parameter model, the 
#    parameter alpha must be strictly greater than 1; all other 
#    parameters are strictly positive (and parameter 3 is 
#    constrained to 1 for an exponential). To avoid problems at 
#    boundaries, set the lower bounds just a wee bit away from them.
#    Set all upper bounds to infinity.
#
lower <- rep(1e-6,length(theta)); lower[3] <- 1; lower[4] <- 1+1e-6
upper <- rep(Inf,length(theta)); upper[3] <- 1
#
# e) Define some storage for the estimates and standard errors in 
#    each month.
#
theta.est <- data.frame(matrix(nrow=12,ncol=length(theta)))
names(theta.est) <- names(theta)
row.names(theta.est) <- month.abb
theta.se <- theta.est
##########################################################################
# 2.	Now, for each month, fit the model using mostly default
#       settings in the auto.fit routine. Looping over months,
#	we need to:
#
#	a)	Extract the appropriate months' data from the 
#		elm.data array, and convert these to an
#		average set of properties and a covariance
#		matrix.
#     b)    Use the estimated variances of the fitting properties
#           to set weights for them in the objective function
#	c)	Call auto.fit() to do the fitting. To keep track
#		of progress, define a graphics window with 
#		room for 8 plots, which will be used to show
#		the best parameter sets found as it iterates,
#		along with their associated objective function 
#		values.
#	d)	Extract the estimates and standard errors
#		from the resulting object, and store these
#
# NB IF YOU'RE PASTING COMMANDS INTO THE R WINDOW, YOU WILL 
#    NEED TO PASTE THE ENTIRE LOOP BEFORE IT STARTS TO RUN.
#
# NB ALSO: THE FITTING IS QUITE TIME-CONSUMING. IF YOU PREFER
# JUST TO FIT A COUPLE OF MONTHS, CHANGE THE RANGE IN THE 
# NEXT LINE - E.G. TO DO JANUARY AND JULY, CHANGE IT TO
#
# for (m in c(1,7)) {
##########################################################################
for (m in 1:12) {
 cat(paste("Starting work on month",m,"...\n"))
#
# a) Extract data for this month; calculate mean and covariance 
#    matrix of fitting properties. NB (i) discard first 2
#    columns of the data array (Month and Year) (ii) there
#    are missing values so we need to specifically exclude
#    these from the calculations (na.rm=TRUE) (iii) the
#    covariance matrix needs to be divided by the number
#    of years, to give the covariance matrix of the *estimates*
#    of the fitting properties.
#
 this.month <- (elm.data$Month == m)
 observed <- colMeans(elm.data[this.month,-c(1,2)],na.rm=TRUE)
 covmat <- var(elm.data[this.month,-c(1,2)],na.rm=TRUE)
 nyears <- sum(!is.na(elm.data[this.month,3]))
 covmat <- covmat/nyears
 w <- 1 / diag(covmat)
#
# b) Define a new graphics array; do the fitting, and store 
#    in an object "fits".
#
 par(mfrow=c(2,4))
 fits <- auto.fit(model.name="BL1",theta=theta,timescales=timescales,
                  observed=observed,covmat=covmat,w=w,lb=lower,ub=upper,
                  log.pars=TRUE)
#
# c) Store the estimates and standard errors
#
 theta.est[m,] <- fits$Theta
 theta.se[m,] <- fits$Std.err
}
##########################################################################
# 3.	Plot the seasonal variation, with fancy plot titles.
#	For each parameter (except parameter 3, which we
#       fixed equal to 1): (i) calculate approximate 95%
#	confidence intervals (+/- 2 standard errors) (ii)
#	plot the estimates with a solid line (iii) add 
#	dashed lines for the confidence intervals (iv) 
#	when all done, copy to both postscript and PDF 
#	files for printing/export
##########################################################################
par(mfrow=c(2,3))
plot.titles <- parse(text=names(theta))
for (p in (1:length(theta))[-3]) {
 ll.95 <- exp(theta.est[,p] - (2*theta.se[,p]))
 ul.95 <- exp(theta.est[,p] + (2*theta.se[,p]))
 ymax <- max(ul.95,na.rm=TRUE); if (ymax > 100) ymax <- 100
 plot(1:12,exp(theta.est[,p]),type="l",lwd=2,xlab="Month",
	ylab="Estimate",ylim=c(0,ymax))
 lines(1:12,ll.95,lwd=2,lty=2)
 lines(1:12,ul.95,lwd=2,lty=2)
 title(main="Seasonal variation in")
 title(main=plot.titles[p],line=1)
}
dev.copy(postscript,"elmfit.ps",paper="A4",horizontal=TRUE)
dev.off()
dev.copy(pdf,"elmfit.pdf",width=11,height=8)
dev.off()

