#############################################################
#
# Routines for fitting point process rainfall models using 
# a method of moments. Not yet documented fully. 
#
# Authors: Georgios Lourmas, (GL), Richard Chandler (REC)
#          and Joao Jesus (JJ)
#
# For update history, see manual.
#
#  UCL 2004-2010
#
#############################################################

#############################################################
# POISSON MODEL
#############################################################
props.PO <- function(theta,timescales) {
 lambda<-theta[1]; mux<-theta[2]; sigmax<- (theta[3]*theta[2]); mul<-theta[4]
 nux<- 1/theta[3]^2 ##added JJ nov 2008
 eta<-1/mul
 h <- timescales$Mean
 EY <- as.numeric(h*lambda*mux*mul)
 h<-timescales$Var
 if (!is.null(h)) {
  VAR<-2*lambda/(eta^3)*(mux^2+sigmax^2)*(eta*h-1+exp(-eta*h))
 } else VAR<-NULL
 h<-timescales$Ph
 if (!is.null(h)) {
  PH <-exp(-lambda*mul)*exp(-lambda*h)
 } else PH <- NULL
 if (!is.null(timescales$Ac)) {
  Ac.row <- 1:(dim(timescales$Ac)[1])
  hh <- timescales$Ac[Ac.row,1]
  kk <- timescales$Ac[Ac.row,2]
#
# Cov and Var both contain terms lambda/(eta^3)*(mux^2+sigmax^2),
# which cancel so no need to evaluate
#
  etah <- -eta*hh
  COV <- exp(etah*(kk-1))+exp(etah*(kk+1))-(2*exp(etah*kk))
  V <- 2*(eta*hh-1+exp(-eta*hh))
   AK<-COV / V
 } else AK <- NULL
##added JJ Nov 2008
 h <- timescales$Skew

## If timescales for Var and Skew are different, the "expected" variance 
## for every timescale in skew needs to be calculated
###
 if (!is.null(h)) {
  var.for.sk <- rep(0,length(h))
#
# JJ: check if variance was calculated for each of the timescales 
# in "Skew". If not calculated, calculate, otherwise use the 
# position indicator to put it in the "correct" position
#
  calculated.vars <- match(h,timescales$Var,0)
  for (i in 1:length(h)){
   if (calculated.vars[i] == 0){
    var.for.sk[i] <- 2*lambda/(eta^3)*(mux^2+sigmax^2)*(eta*h[i]-1+exp(-eta*h[i]))
   } else {
    var.for.sk[i]=VAR[calculated.vars[i]]
   }
  }
  etah <- eta*h
  mom3x <- (nux*(nux+1)*(nux+2))/((nux/mux)^3)
  mu3y <- 6*lambda*mom3x*(etah-2+etah*exp(-etah)+2*exp(-etah))/(eta^4)
  SK = mu3y/var.for.sk^(3/2)
 } else SK <- NULL
 list(Mean=EY,Var=VAR,Ac=AK,Ph=1-PH,Skew=SK)
}

#############################################################
# BARTLETT- LEWIS ORIGINAL MODEL
#############################################################
props.BL0<-function(theta,timescales) {
 lambda <- theta[1]; mux <- theta[2]; sigmax <- theta[3]*mux
 beta <- theta[4]; gamma <- theta[5]; eta <- theta[6]
 kappa <- beta/eta; phi <- gamma/eta; muc <- 1+kappa/phi
# 
# Expected storm duration is (1/(phi*eta)) + ((phi/eta)* double integral)
# (see report 8, page 7). NB this expression is *incorrect* in
# Rodriguez-Iturbe et al. 1987.
#
 mut <- ((1/phi) + (phi*inter2(kappa,phi)) ) / eta
 mur <- (kappa+phi)/(mut*eta*phi)
 ro <- lambda/eta
 mux2 <- mux^2 + sigmax^2
 Gp <- exp(-kappa) * inter(kappa,phi) / eta
 h <- timescales$Mean
 EY <- as.numeric(h*mux*muc*ro)
 h<-timescales$Var
 if (!is.null(h)) {
#
# Original expression by GL. I think this is incorrect - need
# to check (REC)
#
#   VAR <- 2*ro*muc*(2*mux+beta*mux^2/gamma)*h/eta +
#		2*ro*muc*mux^2*beta*eta*(1-exp(-gamma*h))/
#                  (gamma^2*(gamma^2-eta^2)) -
#		2*ro*muc*(2*mux+beta*gamma*mux^2/
#                          (gamma^2-eta^2))*(1-exp(-h*eta))/eta^2
   VAR <- 2*ro*muc*( ( (mux2 + (beta*mux^2/gamma))*h/eta ) +
                     ( mux^2*beta*eta*(1-exp(-gamma*h))/
                       (gamma^2*(gamma^2-eta^2)) ) -
                     ( (mux2 + ( beta*gamma*mux^2/(gamma^2-eta^2) ) ) *
                       (1-exp(-h*eta))/eta^2 ) )
 } else VAR <- NULL
 h<-timescales$Ph
 if (!is.null(h)) {
  PH <-exp(-lambda*(h+mut) +
           (lambda*Gp*(gamma+(beta*exp(-(beta+gamma)*h)))/(beta+gamma))) 
 } else PH <- NULL
 if (!is.null(timescales$Ac)) {
  Ac.row <- 1:(dim(timescales$Ac)[1])
  hh <- timescales$Ac[Ac.row,1]
  kk <- timescales$Ac[Ac.row,2]
  COV <- ro*muc* ( ( (mux2 + (beta*gamma*mux^2/(gamma^2-eta^2)) ) *
                    (1-exp(-eta*hh))^2 * exp(-eta*(kk-1)*hh) / eta^2 ) -
                   (mux^2*beta*eta*(1-exp(-gamma*hh))^2 *
                    exp(-gamma*(kk-1)*hh) / (gamma^2*(gamma^2-eta^2)) ) )
  V <- 2*ro*muc*( ( (mux2 + (beta*mux^2/gamma))*hh/eta ) +
                     ( mux^2*beta*eta*(1-exp(-gamma*hh))/
                       (gamma^2*(gamma^2-eta^2)) ) -
                     ( (mux2 + ( beta*gamma*mux^2/(gamma^2-eta^2) ) ) *
                       (1-exp(-hh*eta))/eta^2 ) )
   AK<-COV / V
 } else AK <- NULL 
##added JJ Nov 2008
 h <- timescales$Skew
 if (!is.null(h)) {
  stop("Skewness not yet implemented for this choice of model") 
 } else SK <- NULL
 list(Mean=EY,Var=VAR,Ac=AK,Ph=1-PH,Skew=SK)
}
####################################################
# Function inter is used for the estimation of the integral used in Pdry
####################################################
inter <-function(kappa,phi) {
 M <- 5
 j <- 0:M
 inter.pdry <-sum(kappa^j/(gamma(j+1)*(1+j+phi)*(j+phi)))
 inter.pdry
}
#######################################################
# Function integ is used for the estimation of the integral used in mut
######################################################
inter2 <-function(kappa,phi) {
 M <- 30
 j <-1:M
 inter.mut <- 1/phi + sum((-kappa)^(j-1)*(kappa-j^2-j)*
                                    exp(lgamma(phi)-lgamma(j+1+phi)) / 
					                     (j*(j+1)))
 inter.mut
}
#############################################################
# BARTLETT- LEWIS RANDOMIZED MODEL
#############################################################
props.BL1 <- function(theta,timescales) {
 lambda <- theta[1]; mux <- theta[2]; sigmax <- theta[3]*theta[2]
 alpha <- theta[4]; nu <-theta[4]/theta[5]; kappa<-theta[6]; phi<-theta[7]
 EXsq <-mux^2+sigmax^2
 muc<-1+kappa/phi
 if (alpha <= 1) {
#
# Use cat() for error printing, because otherwise it doesn't appear
# when called within a try() environment (e.g. during optimisation)
#
  cat(paste("ERROR in props.BL1: alpha must be strictly > 1",
            "(current value is",round(alpha,5),")\n"))
  stop(call.=FALSE)
 }
 expnt <- alpha-3
 if (expnt == 0) expnt <- 1e-8
 h <- timescales$Mean
 EY <- lambda*h*nu*mux*muc/(alpha-1)
 A1 <- (lambda*muc/((alpha-1)*(alpha-2)*(expnt)))*
			(EXsq+kappa*phi*mux^2/(phi^2-1))
 A2 <- (lambda*muc*kappa*mux^2) / 
 			(phi^2*(phi^2-1)*(alpha-1)*(alpha-2)*(expnt))
 mut <- nu * (phi*inter2(kappa,phi)+phi^(-1)) / (alpha-1)
 h<-timescales$Var
 if (!is.null(h)) {
 VAR <- 2*A1*(expnt*h*nu^2 - nu^(3) + ( ((nu+h)^3) *
                                       ((nu/(nu+h))^alpha) )) - 
	2*A2*(phi*expnt*h*nu^2 - nu^3 + ( ((nu+(phi*h))^3)*
                                         ((nu/(nu+(phi*h)))^alpha) ))
 } else VAR <- NULL
 h<-timescales$Ph
 if (!is.null(h)) {
 PH <- exp(-lambda*(h+mut)+lambda*nu*exp(-kappa)/
           (alpha-1)*(phi+kappa*(nu/(nu+(kappa+phi)*h))^(alpha-1))/
           (phi+kappa)*inter(kappa,phi))
 } else PH <- NULL
 if (!is.null(timescales$Ac)) {
 Ac.row<-1:(dim(timescales$Ac)[1])
 hh <- timescales$Ac[Ac.row,1] 
 kk <- timescales$Ac[Ac.row,2]
 COV <-  A1*( ( (nu+(kk+1)*hh)^3 * ((nu /(nu+(kk+1)*hh))^alpha) ) - 
              2*( (nu+kk*hh)^3 * ((nu / (nu+kk*hh))^alpha )) + 
         ( (nu+(kk-1)*hh)^3 * ((nu / (nu+(kk-1)*hh))^alpha )) ) - 
         A2*( ( (nu+(kk+1)*phi*hh)^3 * ((nu / (nu+(kk+1)*phi*hh))^alpha) ) -
              2*( (nu+kk*phi*hh)^3 * ((nu / (nu+(kk*phi*hh)))^alpha) )  + 
         ( (nu+(kk-1)*phi*hh)^3 * ((nu / (nu+(kk-1)*phi*hh))^alpha) ))
 V <- 2*A1*(expnt*hh*nu^2 - nu^(3) + ( ((nu+hh)^3) *
                                     ((nu/(nu+hh))^alpha) )) - 
      2*A2*(phi*expnt*hh*nu^2 - nu^3 + ( ((nu+(phi*hh))^3)*
                                         ((nu/(nu+(phi*hh)))^alpha) ))
 AK<-COV / V
 } else AK <- NULL 
##added JJ Nov 2008
 h <- timescales$Skew
 if (!is.null(h)) {
  stop("Skewness not yet implemented for this choice of model") 
 } else SK <- NULL
 list(Mean=EY,Var=VAR,Ac=AK,Ph=1-PH,Skew=SK)
}

#############################################################
# 2- CELL TYPE MODEL
#############################################################
props.BL2 <-function(theta,timescales) {
 lambda <-theta[1]; mux1 <-theta[2]; mux2 <-theta[2]+theta[3]
 muxsq1 <-(theta[2])^2+(theta[4])^2
 muxsq2 <-(theta[2]+theta[3])^2+(theta[5])^2
 eta1 <-1/theta[6]; eta2 <-1/theta[7]; psi1 <-theta[8]
 muc <- theta[9]; deltas <- theta[10]
 psi2 <- 1-psi1
 if (psi1 < 0) {
#
# Use cat() for error printing, because otherwise it doesn't appear
# when called within a try() environment (e.g. during optimisation)
#
  cat(paste("ERROR in props.BL2: psi1 must be >= 0",
            "(current value is",round(psi1,5),")\n"))
  stop(call.=FALSE)
 }
 if (psi1 > 1) {
  cat(paste("ERROR in props.BL2: psi1 must be <= 1",
            "(current value is",round(psi1,5),")\n"))
  stop(call.=FALSE)
 }
 gamma <- 1/deltas; beta <- (muc-1)/deltas
 C1 <-lambda*muc*(psi1*muxsq1/eta1+2*beta*gamma*psi1*mux1/
                  (gamma^2-eta1^2)*(psi1*mux1/(eta1+eta1)+
                                    psi1*mux1/(eta2+eta1)))
 C2 <-lambda*muc*(psi2*muxsq2/eta2+2*beta*gamma*psi2*mux2/
                  (gamma^2-eta2^2)*(psi2*mux2/(eta1+eta2)+
                                    psi2*mux2/(eta2+eta2)))
 D <-lambda*muc*beta*(psi1^2*mux1^2/((eta1-gamma)*(eta1+gamma))+
                      psi1*psi2*mux1*mux2/((eta1-gamma)*(eta2+gamma))+
	psi2*psi1*mux2*mux1/((eta2-gamma)*(eta1+gamma))+
                      psi2^2*mux2^2/((eta2-gamma)*(eta2+gamma)))
 h <- timescales$Mean
 EY <-as.numeric(h*lambda*muc*(psi1*mux1/eta1+psi2*mux2/eta2))
 h<-timescales$Var
 if (!is.null(h)) {
 VAR <-2*C1/eta1^2*(h*eta1+exp(-eta1*h)-1)+2*C2/eta2^2*
   (h*eta2+exp(-eta2*h)-1)+2*D/gamma^2*(gamma*h+exp(-gamma*h)-1)
 } else VAR <- NULL
 h<-timescales$Ph
 if (!is.null(h)) {
 eta <-psi1*eta1+psi2*eta2
 kappa <- beta/eta; phi <- gamma/eta
 mut <- 1/gamma+gamma/eta^2*inter2(kappa,phi)
 PH <-exp(-lambda*(h+mut)+lambda*(eta^(-1)*
                                  exp(-kappa)*inter(kappa,phi))*
          (gamma+beta*exp(-(beta+gamma)*h)/(beta+gamma)))
 } else PH <- NULL
 if (!is.null(timescales$Ac)) {
 Ac.row<-1:(dim(timescales$Ac)[1])
 hh <- timescales$Ac[Ac.row,1]
 kk <- timescales$Ac[Ac.row,2]
 COV <- C1*exp(-eta1*(kk-1)*hh)/eta1^2*(1-exp(-eta1*hh))^2 +
        C2*exp(-eta2*(kk-1)*hh)/eta2^2*(1-exp(-eta2*hh))^2 +
	D*exp(-gamma*(kk-1)*hh)/gamma^2*(1-exp(-gamma*hh))^2
 V <- 2*C1/eta1^2*(hh*eta1+exp(-eta1*hh)-1) +
      2*C2/eta2^2*(hh*eta2+exp(-eta2*hh)-1)+
      2*D/gamma^2*(gamma*hh+exp(-gamma*hh)-1)
 AK <- COV/V
 } else AK <- NULL 
##added JJ Nov 2008
 h <- timescales$Skew
 if (!is.null(h)) {
  stop("Skewness not yet implemented for this choice of model") 
 } else SK <- NULL
 list(Mean=EY,Var=VAR,Ac=AK,Ph=1-PH,Skew=SK)
}

#############################################################
# BARLETT-LEWIS DEPENDENT DEPTH-DURATION MODEL
#############################################################
props.BL3 <-function(theta,timescales) {
 lambda <-theta[1]; mux <-theta[2];  mux0 <-theta[3]*theta[2]
 deltac <-theta[4]; muc <-theta[5]; deltas <-theta[6]
 c <-1/deltac*(mux0/mux-1);  f <-mux0; beta <-(muc-1)/deltas
 eta <-1/deltac; gamma <-1/deltas; d <-2*c; g <-2*f^2
##############################################################################
## This is the simpified Dependent Depth Duration model (with d=2*c
## and g=2*f^2), thus it is a six parameter model ##
############################################################################## 
 kappa <- beta/eta; phi <- gamma/eta
 mut <- 1/gamma+gamma/eta^2*inter2(kappa,phi)
 dhelp <-d+eta
 chelp <-c+eta
 h <- timescales$Mean
 EY <- as.numeric(lambda*h*muc*eta*f/chelp^2)
 h<-timescales$Var
 if (!is.null(h)) {
 VAR <- 2*lambda*muc*g*eta/dhelp^4*(exp(-dhelp*h)-1+dhelp*h) +
	2*lambda*muc*f^2*beta*eta^2/(chelp^3*(chelp^2-gamma^2))*
          (chelp/gamma^2*(exp(-gamma*h)+gamma*h-1)-gamma/chelp^2*
           (exp(-chelp*h)+chelp*h-1))
 } else VAR <- NULL
 h<-timescales$Ph
 if (!is.null(h)) {
 PH <-exp(-lambda*(h+mut)+lambda*(eta^(-1)*exp(-kappa)*inter(kappa,phi))*
          (gamma+beta*exp(-(beta+gamma)*h)/(beta+gamma)))
 } else PH <- NULL
 if (!is.null(timescales$Ac)) {
 Ac.row<-1:(dim(timescales$Ac)[1])
 hh <- timescales$Ac[Ac.row,1]
 kk <- timescales$Ac[Ac.row,2]
 COV <- lambda*muc*g*eta/dhelp^4*exp(-dhelp*(kk-1)*hh)*
                                    (1-exp(-dhelp*hh))^2 +
 	lambda*muc*f^2*beta*eta^2/(chelp^3*(chelp^2-gamma^2))*
                                    (chelp/gamma^2*(1-exp(-gamma*hh))^2*
        exp(-gamma*(kk-1)*hh)-gamma/chelp^2*(1-exp(-chelp*hh))^2*
                                    exp(-chelp*(kk-1)*hh))
 V <- 2*lambda*muc*g*eta/dhelp^4*(exp(-dhelp*hh)-1+dhelp*hh) +
      2*lambda*muc*f^2*beta*eta^2/(chelp^3*(chelp^2-gamma^2))*
      (chelp/gamma^2*(exp(-gamma*hh)+gamma*hh-1)-gamma/chelp^2*
                                    (exp(-chelp*hh)+chelp*hh-1))
 AK<-COV / V
 } else AK <- NULL
##added JJ Nov 2008
 h <- timescales$Skew
 if (!is.null(h)) {
  stop("Skewness not yet implemented for this choice of model") 
 } else SK <- NULL
 list(Mean=EY,Var=VAR,Ac=AK,Ph=1-PH,Skew=SK)
}
#############################################################
# QUADRATIC RANDOM PARAMETER MODEL
#############################################################
props.QA <-function(theta,timescales) {
 lambda <-theta[1]; mux <-theta[2]; muxsq <-theta[3]; alpha <-theta[4]
 nu <-theta[5]; kappa1 <-theta[6]; kappa2<-theta[7]; phi <-theta[8];
 h <- timescales$Mean
 EY <-as.numeric(lambda*h*mux*(nu/(alpha-1)*(1+kappa1/phi)+kappa2/phi))
 h<-timescales$Var
 if (!is.null(h)) {
 xi1f <-1-(nu/(nu+phi*h))^(alpha-1)
 xi11 <-1-(nu/(nu+1*h))^(alpha-1)
 xi2f <-1-(nu/(nu+phi*h))^(alpha-2)
 xi21 <-1-(nu/(nu+1*h))^(alpha-2)
 xi3f <-1-(nu/(nu+phi*h))^(alpha-3)
 xi31 <-1-(nu/(nu+1*h))^(alpha-3)
 VAR <-2*lambda*h*kappa2^2*mux^2/phi^2+2*lambda*nu*kappa2/((alpha-1)*phi)*
   (h*muxsq+h*mux^2*(1+2*kappa1/phi)+
	mux^2*kappa2/(phi^2*(phi^2-1))*xi1f-kappa2*mux^2*phi*xi11/(phi^2-1))+
	2*lambda*nu^2/((alpha-1)*(alpha-2))*((1+kappa1/phi)*h*
                                             (muxsq+mux^2*kappa1/phi)+
	kappa2*mux^2/(phi^2*(phi^2-1))*(1+2*kappa1/phi)*xi2f-
	kappa2*xi21*(muxsq/phi+mux^2*phi/(phi^2-1)*(1+2*kappa1/phi)))+
	2*lambda*nu^3*(1+kappa1/phi)/((alpha-1)*(alpha-2)*(alpha-3))*
          (mux^2*kappa1*xi3f/(phi^2*(phi^2-1))-
	(muxsq+mux^2*phi*kappa1/(phi^2-1))*xi31)
 } else VAR <- NULL
 if (!is.null(timescales$Ac)) {
 Ac.row<-1:(dim(timescales$Ac)[1])
 hh <- timescales$Ac[Ac.row,1]
 kk <- timescales$Ac[Ac.row,2]
 A11 <- nu/(alpha-1)*((nu/(nu+1*(kk-1)*hh))^(alpha-1)-
        2*(nu/(nu+1*kk*hh))^(alpha-1)+(nu/(nu+1*(kk+1)*hh))^(alpha-1))
 A1f <- nu/(alpha-1)*((nu/(nu+phi*(kk-1)*hh))^(alpha-1)-
        2*(nu/(nu+phi*kk*hh))^(alpha-1)+(nu/(nu+phi*(kk+1)*hh))^(alpha-1))
 A21 <- nu^2/((alpha-1)*(alpha-2))*((nu/(nu+1*(kk-1)*hh))^(alpha-2)-
        2*(nu/(nu+1*kk*hh))^(alpha-2)+(nu/(nu+1*(kk+1)*hh))^(alpha-2))
 A2f <- nu^2/((alpha-1)*(alpha-2))*((nu/(nu+phi*(kk-1)*hh))^(alpha-2)-
        2*(nu/(nu+phi*kk*hh))^(alpha-2)+(nu/(nu+phi*(kk+1)*hh))^(alpha-2))
 A31 <- nu^3/((alpha-1)*(alpha-2)*(alpha-3))*((nu/(nu+1*(kk-1)*hh))^(alpha-3)-
        2*(nu/(nu+1*kk*hh))^(alpha-3)+(nu/(nu+1*(kk+1)*hh))^(alpha-3))
 A3f <- nu^3/((alpha-1)*(alpha-2)*(alpha-3))*((nu/(nu+phi*(kk-1)*hh))^(alpha-3)-
        2*(nu/(nu+phi*kk*hh))^(alpha-3)+(nu/(nu+phi*(kk+1)*hh))^(alpha-3))
 COV <- lambda*(1+kappa1/phi)*(muxsq+kappa1*phi*mux^2/(phi^2-1))*A31+
	lambda*kappa2*(muxsq/phi+phi*(1+2*kappa1/phi)/(phi^2-1)*mux^2)*A21+
        lambda*kappa2^2*mux^2/(phi^2-1)*A11-
	lambda*mux^2/(phi^2*(phi^2-1))*((1+kappa1/phi)*kappa1*A3f+
                                        (1+2*kappa1/phi)*kappa2*A2f+
                                        kappa2^2/phi*A1f)
 xi1f <-1-(nu/(nu+phi*hh))^(alpha-1)
 xi11 <-1-(nu/(nu+1*hh))^(alpha-1)
 xi2f <-1-(nu/(nu+phi*hh))^(alpha-2)
 xi21 <-1-(nu/(nu+1*hh))^(alpha-2)
 xi3f <-1-(nu/(nu+phi*hh))^(alpha-3)
 xi31 <-1-(nu/(nu+1*hh))^(alpha-3)
 V <- 2*lambda*hh*kappa2^2*mux^2/phi^2+2*lambda*nu*kappa2/((alpha-1)*phi)*
       (hh*muxsq+hh*mux^2*(1+2*kappa1/phi)+
	  mux^2*kappa2/(phi^2*(phi^2-1))*xi1f-kappa2*mux^2*phi*xi11/(phi^2-1))+
	  2*lambda*nu^2/((alpha-1)*(alpha-2))*((1+kappa1/phi)*hh*
                                               (muxsq+mux^2*kappa1/phi)+
	  kappa2*mux^2/(phi^2*(phi^2-1))*(1+2*kappa1/phi)*xi2f-
	  kappa2*xi21*(muxsq/phi+mux^2*phi/(phi^2-1)*(1+2*kappa1/phi)))+
	  2*lambda*nu^3*(1+kappa1/phi)/((alpha-1)*(alpha-2)*(alpha-3))*
              (mux^2*kappa1*xi3f/(phi^2*(phi^2-1))-
	  (muxsq+mux^2*phi*kappa1/(phi^2-1))*xi31)
 AK <-COV/V
 } else AK <- NULL
 h <- timescales$Ph
 if (!is.null(h)) {
  cat("Pdry not yet implemented for model QA\n")
  stop(call.=FALSE)
 } else PH <- NULL
##added JJ Nov 2008
 h <- timescales$Skew
 if (!is.null(h)) {
  stop("Skewness not yet implemented for model QA") 
 } else SK <- NULL
 list(Mean=EY,Var=VAR,Ac=AK,Ph=1-PH,Skew=SK)
}
#############################################################
# NEYMAN-SCOTT WITH POISSON DISTRIBUTION FOR NUMBER OF CELLS MODEL.
# EXPRESSIONS TAKEN FROM COWPERTWAIT 1991 AND ENTEKHABI ET AL 1989; 
# CHECKED AGAINST RESULTS GIVEN BY ENTKHABI ET AL (EXCEPT PDRY)
#############################################################
props.NS0 <-function(theta,timescales) {
 lambda <- theta[1]; mux <- theta[2]; sigmax <- theta[3]*mux
 muc<-theta[4]; beta <- theta[5]; eta <- theta[6]
 ro <-lambda/eta; mux2 <- mux^2 + sigmax^2
#
# Next line is E[C(C-1)] where C is number of cells in a storm. The 
# expression given is for the case when C-1 is Poisson distributed
# with mean muc-1 (this is the case considered by Cowpertwait). To
# reproduce the results of Entekhabhi et al, set next line to
# muc^2 + (2*muc)
#
 EC <-muc^2 - 1
 h <- timescales$Mean
 EY <- as.numeric(ro*muc*mux*h)
 h <- timescales$Var
 if (!is.null(h)) {
 A <- (eta*h) - 1 + exp(-eta*h)
 B <- (beta*h) - 1 + exp(-beta*h)
 VAR <- EC*mux^2*( ((beta^3)*A) - ((eta^3)*B) ) /
                 (beta*((beta^2)-(eta^2))) + (2*muc*mux2*A)
 VAR <- VAR*lambda/(eta^3)
 } else VAR <- NULL
 if (!is.null(timescales$Ac)) {
  Ac.row <- 1:(dim(timescales$Ac)[1])
  hh <- timescales$Ac[Ac.row,1] 
  kk <- timescales$Ac[Ac.row,2]
  A1 <- (eta*hh) - 1 + exp(-eta*hh)
  B1 <- (beta*hh) - 1 + exp(-beta*hh)
  A2 <- ( (1-exp(-eta*hh))^2 ) * exp(-eta*hh*(kk-1)) / 2
  B2 <- ( (1-exp(-beta*hh))^2 ) * exp(-beta*hh*(kk-1)) / 2
  COV <- EC*(mux^2)*( ((beta^3)*A2) - ((eta^3)*B2) ) /
                    (beta*((beta^2)-(eta^2))) + (2*muc*mux2*A2)
  COV <- COV*lambda/(eta^3)
  V <- EC*(mux^2)*( ((beta^3)*A1) - ((eta^3)*B1) ) /
                  (beta*((beta^2)-(eta^2))) + (2*muc*mux2*A1)
  V <- V*lambda/(eta^3)
  AK <- COV / V
 } else AK <- NULL
 h <- timescales$Ph
 if (!is.null(h)) {
#
# Need to use numerical integration for Pdry; integrate() needs a scalar
# argument so we'll have to loop
#
  PH <- rep(NA,length(h))
#
#	JJ:
#	mu = E(C-1)
#	muc = E(C)
#	mu = muc - 1 , instead of mu = muc +1, Comment above says this has not been tested
#
	mu <- muc - 1
#mu <- muc +1
  for (i in 1:length(PH)) {
   pd <- h[i] - (1 - exp( mu*(exp(-beta*h[i])-1) ))/(beta*mu) +
                integrate(pt.ns0,0,Inf,beta=beta,eta=eta,mu=mu,h=h[i],
                          stop.on.error=FALSE)$value
   PH[i] <- exp(-lambda*pd)
  }
 } else PH <- NULL
##added JJ Nov 2008
 h <- timescales$Skew
 if (!is.null(h)) {
  stop("Skewness not yet implemented for this choice of model") 
 } else SK <- NULL
 list(Mean=EY,Var=VAR,Ac=AK,Ph=1-PH,Skew=SK)
}
#############################################################
# Here's the function to be integrated to get Pdry for NS model. The
# function is actually (1-pt(h)) in Cowpertwait's notation, since
# this is the required integrand.
#############################################################
pt.ns0 <- function(t,beta,eta,mu,h) {
 e1 <- exp(-beta*t); e2 <- exp(-eta*t)
 term1 <- exp(-beta*(t+h)) + 1 - ( ((eta*e1) - (beta*e2)) / (eta-beta) )
 term2 <- beta*(e1-e2)/(eta-beta) + e1 - exp(-beta*(t+h))
 term2 <- exp(-mu*term2)
 1 - (term1*term2)
}
#############################################################
# RANDOM PARAMETER NEYMAN-SCOTT MODEL. EXPRESSIONS TAKEN FROM 
# ENTEKHABI ET AL 1989; CHECKED AGAINST RESULTS GIVEN IN THAT
# PAPER. NB HOWEVER TAKE C-1 TO HAVE A POISSON DISTRIBUTION
# FOLLOWING COWPERTWAIT (C BEING THE NUMBER OF CELLS PER STORM)
#
# NB I HAVE MY DOUBTS AS TO WHETHER THE FIRST TERM IN THE 
# EXPRESSION FOR VARIANCE SHOULD BE THERE - IT LOOKS AS THOUGH
# THE EXPRESSION IS FOR E[Y^2] RATHER THAN VAR[Y]. THERE ARE
# 2 LINES COMMENTED OUT BELOW, GIVING WHAT I THINK MAY BE THE
# CORRECT EXPRESSIONS. ALSO, NB I THINK THE EXPRESSIONS 
# GIVEN BY ENTEKHABI ARE BASED ON TAYLOR SERIES EXPANSIONS OF
# (beta^2-eta^2)^-1, VALID WHEN BETA^2 << ETA^2.
#############################################################
props.NS1 <-function(theta,timescales) {
 lambda <- theta[1]; mux <- theta[2]; sigmax <- theta[3]*theta[2]
 alpha <- theta[4]; nu <-theta[4]/theta[5]; muc<-theta[6]; beta<-theta[7]
 if (alpha <= 1) {
#
# Use cat() for error printing, because otherwise it doesn't appear
# when called within a try() environment (e.g. during optimisation)
#
  cat(paste("ERROR in props.NS1: alpha must be strictly > 1",
            "(current value is",round(alpha,5),")\n"))
  stop(call.=FALSE)
 }
#
# See comment in props.NS0 regarding E[C(C-1)] 
#
 mux2 <- mux^2 + sigmax^2; EC <- muc^2 - 1
 I1 <- I.NS1(1,0,alpha,nu); I2 <- I.NS1(2,0,alpha,nu)
 muy <- lambda*mux*muc*I1
 h <- timescales$Mean
 EY <- as.numeric(muy*h)
 C1 <- lambda*muc*mux2; C2 <- lambda*EC*((mux*beta)^2)
 h <- timescales$Var
 if (!is.null(h)) {
 VAR <- ((muy*h)^2) + ( (2*C1*h) + 
# VAR <- ( (2*C1*h) + 
                        (C2*((beta*h)+exp(-beta*h)-1)/(beta^3)) +
                       (lambda*mux*muc*h)^2 )*I2 -
         (2*C1*I.NS1(3,0,alpha,nu)) - (C2*h*I.NS1(4,0,alpha,nu)) +
         (C2*I.NS1(5,0,alpha,nu)) + (2*C1*I.NS1(3,h,alpha,nu)) -
         (C2*I.NS1(5,h,alpha,nu))
 } else VAR <- NULL
 if (!is.null(timescales$Ac)) {
  Ac.row <- 1:(dim(timescales$Ac)[1])
  hh <- timescales$Ac[Ac.row,1] 
  kk <- timescales$Ac[Ac.row,2]
  COV <- (C1*I.NS1(3,(kk-1)*hh,alpha,nu)) + 
         (C1*I.NS1(3,(kk+1)*hh,alpha,nu)) - 
         (2*C1*I.NS1(3,kk*hh,alpha,nu)) -
         (C2*I.NS1(5,(kk-1)*hh,alpha,nu)/2) - 
         (C2*I.NS1(5,(kk+1)*hh,alpha,nu)/2) + 
         (C2*I.NS1(5,kk*hh,alpha,nu)) +
         (C2*((1-exp(-beta*hh))^2)*exp(-beta*(kk-1)*hh)*I2/(2*beta^3)) +
         ((muc*mux*lambda*hh)^2)*(I2-(I1^2))
  V <- ((muy*hh)^2) + ( (2*C1*hh) + 
#  V <-  ( (2*C1*hh) + 
                        (C2*((beta*hh)+exp(-beta*hh)-1)/(beta^3)) +
                   (lambda*mux*muc*hh)^2 )*I2 -
       (2*C1*I.NS1(3,0,alpha,nu)) - (C2*hh*I.NS1(4,0,alpha,nu)) +
       (C2*I.NS1(5,0,alpha,nu)) + (2*C1*I.NS1(3,hh,alpha,nu)) -
       (C2*I.NS1(5,hh,alpha,nu))
  AK <- COV / V
 } else AK <- NULL
 h <- timescales$Ph
 if (!is.null(h)) {
  cat("Pdry not yet implemented for model NS1\n")
  stop(call.=FALSE)
 } else PH <- NULL
##added JJ Nov 2008
 h <- timescales$Skew
 if (!is.null(h)) {
  stop("Skewness not yet implemented for this choice of model") 
 } else SK <- NULL
 list(Mean=EY,Var=VAR,Ac=AK,Ph=1-PH,Skew=SK)
}

#############################################################
# Utility function for NS1 properties, as defined in 
# Entekhabi et al. In that paper, there's a factor of
# gamma(alpha-x)/gamma(alpha), which simplifies to 
# 1/[(alpha-1)...(alpha-x)] for integer x (computationally cheaper)
#############################################################
I.NS1 <- function(x,y,alpha,nu) {
 if (x%% 1 != 0) stop("x must be an integer")
 (nu^alpha) * (nu+y)^(x-alpha) / prod(alpha-(1:x))
}
#########################################################################
#  Choice of model for which to compute properties
#########################################################################
model.select <-function(model.name,theta,timescales,log.props) {
 if (model.name==c("PO")) {
  expected <-props.PO(theta,timescales)
 }
 if (model.name==c("BL1")) {
  expected <-props.BL1(theta,timescales)
 }
 if (model.name==c("BL0")) {
  expected <-props.BL0(theta,timescales)
 }
 if (model.name==c("BL3")) {
  expected <- props.BL3(theta,timescales)
 }
 if (model.name==c("NS0")) {
  expected<-props.NS0(theta,timescales)
 }
 if (model.name==c("NS1")) {
  expected<-props.NS1(theta,timescales)
 }
  if (model.name==c("BL2")) {
  expected<-props.BL2(theta,timescales)
 }
  if (model.name==c("QA")) {
  expected<-props.QA(theta,timescales)
 }
 expected <- unlist(expected)
 if (sum(log.props) > 0) expected[log.props] <- log(expected[log.props])
 expected
} 
#########################################################################
#  Almost the same again, but with arguments in a different order
#  to allow numerical differentiation. Also accomodates inverse
#  log transformation, in case gradients are required on the log
#  scale
#########################################################################
grad.fn <- function(x,model,timescales,log.pars,log.props) {
 theta <- x; theta[log.pars] <- exp(x[log.pars])
 unlist(model.select(model.name=model,theta=theta,timescales,log.props))
}
#########################################################################
# Objective function for method of moments. This is designed to be
# called from moment.fit and should not be called directly.
#
# Arguments:
#
# p		Parameter vector at which to evaluate the function
#		(possibly transformed after a call to par.trans)
# backtrans Do we need to back-transform the values held in p?
# lb		Lower limits on parameters (original scale before
#		the call to par.trans)
# ub		Upper limits, similarly
# meth	Fitting method being used (passed by moment.fit)
# observed	Observed properties of rainfall sequence
# w		Either a vector of weights to use in fitting, or a weighting
#		matrix
# timescalesList of 5 components (Mean, Var, Ac, Ph and Skew).
#		Each component contains a vector (matrix, in the 
#           case of Ac) of timescales at which to evaluate the 
#           corresponding properties.
# model.nameChoice of model 
# log.pars	Logical scalar or vector indicating parameters
#		that have been log-transformed (in addition to
#		the transformation that may need to be undone
#		using backtrans).
# compute.grad  Logical scalar: return an object with a gradient
#		attribute. This can be used in numerical minimisation
#		using nlm(), for example - the built-in derivatives
#		in nlm() can be inaccurate.
# compute.hess	As compute.grad, but controlling the calculation
#		of a hessian. If TRUE, it is returned as an attribute
#		of the calculated object.
# ...		Additional arguments to num.deriv and num.hess, for
#		use if either grad or hess is TRUE
#########################################################################
objfunc <- function(p,backtrans,lb,ub,observed,w=rep(1,length(observed)),
                    timescales,model.name,log.pars,log.props,
                    compute.grad=FALSE,compute.hess=FALSE,...) {
#
#    nlm() has a nasty habit of passing NAs or infinities from 
#    time to time. Trap these here and send something nasty back!
#
  if (any(!is.finite(p))) return(NA)
  if (backtrans) theta <- par.trans(p,lb,ub,"backward") else theta <- p
  theta[log.pars] <- exp(theta[log.pars])
  expected <- model.select(model.name,theta,timescales,log.props)
  if (is.vector(w)) {
   sumsq <- sum(w*(unlist(observed)-unlist(expected))^2)
  } else if (is.matrix(w)) {
   OminusE <- unlist(observed)-unlist(expected)
   sumsq <- as.vector(t(OminusE) %*% w %*% OminusE)
  }
  if (compute.grad) {
   if (backtrans) {
    lolim <- -Inf; hilim <- Inf
   } else {
    lolim <- lb; hilim <- ub
   }
   attr(sumsq,"gradient") <- num.deriv(objfunc,p,xlo=lolim,xhi=hilim,
				lb=lb,ub=ub,observed=observed,w=w,
				backtrans=backtrans,timescales=timescales,
				model.name=model.name,
				log.pars=log.pars,log.props=log.props,
                                ...)$gradient
  }
  if (compute.hess) {
   if (backtrans) {
    lolim <- -Inf; hilim <- Inf
   } else {
    lolim <- lb; hilim <- ub
   }
   attr(sumsq,"hessian") <- num.hess(objfunc,p,xlo=lolim,xhi=hilim,
				lb=lb,ub=ub,observed=observed,w=w,
				backtrans=backtrans,timescales=timescales,
				model.name=model.name,
				log.pars=log.pars,log.props=log.props,
                                ...)$hessian
  }
  sumsq
 }
#########################################################################
# 	Routine for "automatic" model fitting, with options. The
#       routine starts by carrying out a large number of optimisations
#       from various starting points, within limits that may be set
#       by the user. It then uses the results from these fits to try
#       and home in on likely regions of parameter space for more
#       starting points.
#	
#	Arguments:
#     model.name  The identifier for the model to fit
#	theta		Starting value for parameter vector
#     timescales  A list selecting the fitting statistics
#	observed	Vector of observed statistics
#	covmat	Estimated covariance matrix of observed
#			statistics. This is optional; if present,
#			it may be used to calculate standard errors
#			on parameters.
#	stderr	Logical: should standard errors be computed?
#     numeric.hessLogical: if TRUE, the hessian of the objective
# 			function will be computed numerically. If
#			FALSE, an analytic approximation will be 
#			used. The approximation can be regarded
#			as an alternative estimator of the true
#			expected Hessian, and is valid to the same
#			order of asymptotic approximation as the 
#			numerical version; it has the additional
#			advantage that it is guaranteed positive
#			(semi-)definite by construction. Defaults
#			to FALSE.
#	conf.level  A vector of probabilities from which, if
#			stderr=TRUE, objective function thresholds 
#			will be calculated corresponding to approximate 
#			confidence intervals. The default value of
#			c(0.95,0.99) returns thresholds corresponding
#			to approximate 95% and 99% intervals.
#	w		Either a vector of weights to associate with 
#			observed statistics, or a weighting matrix
#	log.pars	A logical constant or vector indicating whether
#			to optimise over the (natural) logarithms of
#			some parameters. For parameters where this is
#			TRUE, the routine transforms the parameter
#			constraints accordingly, and computes standard
#			errors etc. on the log scale.
#     log.props   Logical constant or vector, the same length as
#                 "observed", indicating whether or not the
#                 properties in "observed" are estimates of the
#                 corresponding properties on the original or
#                 log scale.
#	print.level	Controls amount of output. Values:
#			0	Work silently
#                 1     Print a message each time a new set
#                       of limits is calculated
#			>2	Increase detail from moment.fit
#	lb		Lower bounds on parameters. Defaults to 
#			`almost zero', to avoid problems on the 
#			boundary itself when you divide by a parameter
#	ub		Upper bounds on parameters
#     ninit       Number of initial Nelder-Mead optimisations to
#                 carry out in order to identify likely regions
#                 of the parameter space.
#     nkeep       The number of fits to keep at each stage.
#     nsubseq     Total number of optimisations in table after
#                 initial Nelder-Mead phase.
#     nsearches   Number of nlm() passes (each of which will 
#                 expand the table to contain nsubseq optimisations)
#                 to carry out
#     reltol      Relative tolerance: a solution is deemed to
#                 have been found when the limits on the
#                 parameters are all within reltol of each other.
#	plot.progress	If TRUE, plots are produced after each round
#			of fitting to show where we've got to.
#
#	Value: a list containing the following components:
#  	Theta		The optimal parameter vector found. For 
#			elements to which a log transformation has
#			been applied, this contains the natural 
#			logarithm of the original parameter, and 
#			this is reflected in the names attribute
#			of the result.
#	SE		If requested, a vector of standard errors
#			on the parameter estimates
#	Corr		If SE was requested, the correlation matrix
#			of the parameter estimates
#	Obj     	The value of the objective function at Theta
#	Obj.thresh	If SE was requested, values of the objective
#			function that define approximate confidence
#			regions at the levels specified in conf.level.
#	Gradient	The gradient vector of the objective function
#			at Theta
#     Hessian	The Hessian matrix of the objective function
#			at Theta
#	fits.table	A table containing the final points of 
#			other optimisations whose performance is
#			in some sense close to that of Theta.
#	Bounds	The final constraints used to find the
#			optimum
#	Model		The value of model.name on input
#	Timescales	The value of timescales on input
#	Observed	The value of observed on input
#	w		The weights used in the objective function
#	log.pars	A vector indicating which of the 
#			parameters have had a log transformation
#			applied
#########################################################################
auto.fit <- function(model.name,theta,timescales,observed,covmat=NULL,
		     w=rep(1,length(observed)),log.pars=FALSE,
                 log.props=FALSE,stderr=!is.null(covmat),
                 numeric.hess=FALSE,conf.level=c(0.95,0.99),print.level=1,
                 lb=rep(1e-6,length(theta)),ub=rep(Inf,length(theta)),
                 ninit=100,nkeep=max(1,0.2*ninit),
		     nsubseq=2.5*nkeep,nsearches=5,reltol=0.01,
                 plot.progress=TRUE) {
  lolim <- lb; hilim <- ub; startval <- theta; np <- length(theta)
  colnos <- (2*(1:np)) - 1
  if (nsubseq <= nkeep) stop("You must have nsubseq > nkeep")
  if (any(lb > ub)) stop("Some lower bounds are greater than upper bounds")
  if (length(log.props) > 1 && length(log.props) != length(observed)) {
   stop(paste("If 'log.props' isn't a scalar,",
              "it should be the same length as 'observed'"))
  }
  if (is.matrix(w)) wmat <- w else wmat <- diag(w)

#
#	If the starting value is on a boundary, add a tiny
#	adjustment to circumvent problems
#
  par.adj <- (hilim - lolim)*1e-6
  par.adj[is.infinite(par.adj)] <- 1e-6
  startval[startval == lolim] <- (startval+par.adj)[startval == lolim]
  startval[startval == hilim] <- (startval-par.adj)[startval == hilim]
  for (p in 1:np) {
    if (lb[p] > theta[p]) {
      stop(paste("\nAll starting values must be strictly greater",
                 "than lower bounds -","\nfor parameter",p,
                 "the starting value is",theta[p],
                 "\nbut lower bound is",lb[p]))
    } else if (ub[p] < theta[p]) {
      stop(paste("\nAll starting values must be strictly less",
                 "than upper bounds -","\nfor parameter",p,
                 "the starting value is",theta[p],
                 "\nbut upper bound is",ub[p]))
    }
  }
#
# 	Now deal with log transformation where requested. 
#	Changing lb and ub here shouldn't be a problem, because
#	the corresponding values in the calling function do
#	*not* get changed. R is good like that ...
#
  startval[log.pars] <- log(startval[log.pars])
  lb[log.pars] <- log(lb[log.pars])
  lolim[log.pars] <- log(lolim[log.pars])
  ub[log.pars] <- log(ub[log.pars])
  hilim[log.pars] <- log(hilim[log.pars])
#
# Make some nice names for the parameters
#
  if (is.null(names(theta))) {
   par.names <- paste("Parameter",1:np)
  } else {
   par.names <- names(theta)
  }
  par.names[log.pars] <- paste("log(",par.names[log.pars],")",sep="")
  names(startval) <- par.names
#
# Suppress printing of warning messages
#
  warnings.value <- getOption("warn"); options(warn=-1)
#
#    Initial Nelder-Mead optimisations. NB it's possible that none
#    of these will produce a solution at the first attempt - if
#    this is the case, go back and try again. Also, to save time
#    don't store gradients here.
#
  if (plot.progress) par(mfrow=par("mfrow"))
  limtable <- data.frame(Lower=lolim,Upper=hilim,row.names=par.names)
  if (print.level >= 1) {
    cat("Starting initial search\n")
    cat("Initial bounds on parameters:\n")
    print(limtable)
  }
  init.done <- FALSE
  while (!init.done) {
   if (ninit > 0) {
    fits <- moment.fit(model.name,startval,timescales,observed,w,
                       "Nelder-Mead",ninit,max(0,print.level-1),
                       lolim,hilim,log.pars,log.props,gradients=FALSE,
                       plot.progress)
   } else {
    fits <- moment.fit(model.name,startval,timescales,observed,w,
                       "Nelder-Mead",1,max(0,print.level-1),
                       startval,startval,log.pars,log.props,FALSE,FALSE)
    nfits <- 1
   }
   init.done <- !all(is.na(fits$Obj))
  }
  searchnum <- 1; maxint <- max(2*(hilim-lolim)/abs(hilim+lolim))
  if(is.na(maxint)) maxint <- 1 + reltol
  while (maxint > reltol && searchnum <= nsearches) {
    if (print.level >= 1) cat(paste("Subsequent search",searchnum,"\n"))
    nfits <- dim(fits)[1]
    thresh <- quantile(fits$Obj,probs=min(nkeep/nfits,1),na.rm=TRUE)
    best.fits <- fits$Obj <= thresh
    best.fits[is.na(best.fits)] <- FALSE
#
#    Heuristic: optimal parameter set probably lies roughly within 
#    range of best fits found so far, if these are roughly in the
#    centre of *all* fits currently being considered (which, after
#    search 1, will consist only of fits that yield a pretty good
#    value of the objective function - within a factor of 10 of the
#    worst of the `best' fits). Take `roughly within range of best
#    fits' to mean `add 10% to each end of this range'; and take
#    `roughly in the centre' to mean `between the 10th and 90th
#    percentiles'. If, on the basis of this criterion, we are 
#    *not* able to decide that the optimal parameter set lies within 
#    the range, move the existing limits, to an extent determined
#    by how close the best fits are to the end of the range (if
#    they're a long way away we could tighten the limits a bit).
#    Three caveats: (1) never go outside the range requested by
#    the user (2) never move an infinite limit unless the best fits
#    are all nicely in the middle so you're confident about what
#    you're doing (3) don't do *anything* without at least 5
#    fits to work with.
#
    par.lims <- data.frame(P10=lolim,P90=hilim,LB=lolim,UB=hilim)
    par.lims$P10 <- apply(fits[,colnos],2,quantile,p=0.1,na.rm=TRUE)
    par.lims$P90 <- apply(fits[,colnos],2,quantile,p=0.9,na.rm=TRUE)
    par.lims$LB <- apply(fits[best.fits,colnos],2,min,na.rm=TRUE)
    par.lims$UB <- apply(fits[best.fits,colnos],2,max,na.rm=TRUE)
    tight.lb <- par.lims$LB > par.lims$P10
    tight.ub <- par.lims$UB < par.lims$P90
    par.adj <- (hilim - lolim)
    finlim <- is.finite(lolim)
    if (sum(best.fits) >= 5) {
     lolim[finlim] <- (par.lims$LB -
                       (par.adj*par.adj/
                        (1+par.lims$LB-lolim)))[finlim]
     finlim <- is.finite(hilim)
     hilim[finlim] <- (par.lims$UB +
                       (par.adj*par.adj/
                        (1+hilim-par.lims$UB)))[finlim]
     par.adj <- 0.1*(par.lims$UB-par.lims$LB)
     lolim[tight.lb] <- par.lims$LB[tight.lb]-par.adj[tight.lb]
     hilim[tight.ub] <- par.lims$UB[tight.ub]+par.adj[tight.ub]
     lolim[lolim < lb] <- lb[lolim < lb]
     hilim[hilim > ub] <- ub[hilim > ub]
    }
    maxint <- max(2*(hilim-lolim)/(hilim+lolim))
    if(is.na(maxint)) maxint <- 1 + reltol
    fits <- fits[!is.na(fits$Obj) & fits$Obj <= thresh,]
#
#	Use best fit found to date as starting point next time
#	round. Move boundary points a bit, as before. 
#
    startval <- as.numeric((fits[order(fits$Obj),colnos])[1,])
    par.adj <- (hilim - lolim)*1e-6
    par.adj[is.infinite(par.adj)] <- 1e-6
    startval[startval == lolim] <- (startval+par.adj)[startval == lolim]
    startval[startval == hilim] <- (startval-par.adj)[startval == hilim]
    names(startval) <- par.names
    limtable <- data.frame(Theta=startval,Lower=lolim,Upper=hilim,
                           row.names=par.names)
    if (print.level >= 1) {
      cat("Current optimum:\n")
      print(limtable)
      cat(paste("Objective function value:",
                signif(min(fits$Obj,na.rm=TRUE),4),"\n\n"))
    }
    nfits <- dim(fits)[1]
    fits2 <- moment.fit(model.name,startval,timescales,observed,w,
                       "nlm",nsubseq-nfits,max(0,print.level-1),
                       lolim,hilim,log.pars,log.props,plot=FALSE)
#
#    Only keep values with objective functions better than 10*thresh
#
    fits2 <- fits2[!is.na(fits2$Obj) & fits2$Obj <= 10*thresh,]
    fits <- rbind(fits,fits2)
    if (plot.progress) {
      par(mfrow=par("mfrow"))
      plot.mmfit(fits,pltfrac=1,xlimits=limtable[,2:3],
                 plot.grad=TRUE,nonconverged=TRUE)
    }
    searchnum <- searchnum + 1
    if (maxint <= reltol & print.level >= 1) {
      cat(paste("Relative difference between bounds < reltol (set to ",
                reltol,")\nfor all parameters - ",
                "no further searches needed\n"))
    }
  }
  thresh <- quantile(fits$Obj,probs=min(nkeep/nfits,1),na.rm=TRUE)
  fits <- fits[fits$Obj <= thresh,]

#
#    Call nlm a final time, now using num.deriv to calculate 
#    derivatives if standard errors are required (this seems to be 
#    more accurate than the internal routines, but correspondingly 
#    slower so not worth doing unless serious accuracy is required -
#    which it may be if we're going to calculate standard errors. 
#
  smallest <- order(fits$Obj,na.last=TRUE)[1]
  startval <- as.numeric(fits[smallest,colnos])
  names(startval) <- par.names
  phi <- par.trans(startval,lolim,hilim,"forward")
  gradvec <- as.numeric(fits[smallest,colnos+1])
  thetahat <- try(nlm(objfunc,phi,typsize=abs(phi),
                      fscale=abs(fits$Obj[smallest]),
                      observed=observed,w=w,backtrans=TRUE,
                      timescales=timescales,model.name=model.name,
                      log.pars=log.pars,log.props=log.props,
                      lb=lolim,ub=hilim,compute.grad=stderr,
                      check.analyticals=FALSE,iterlim=200,maxit=3),
                  silent=TRUE)
#
#    Translate parameters and gradient back to original scale
#
  delta <- grad.trans(phi,lolim,hilim)
  if (isTRUE(class(thetahat) == "try-error")) {
    Theta <- startval
    Gradient <- rep(NA,np)
    Obj.func <- min(fits$Obj,na.rm=TRUE)
  } else {
    phi <- thetahat$estimate
    Theta <- par.trans(phi,lolim,hilim,"backward")
    Gradient <- thetahat$gradient * delta
    Obj.func <- thetahat$minimum
  }
  names(Theta) <- par.names
  names(Gradient) <- par.names
  rownames(fits) <- 1:dim(fits)[1]
#
# Reset warnings
#
  options(warn=warnings.value)
#
# Now, if requested (and if the user has supplied covmat), 
# calculate standard errors. Treat parameters for which lower and
# and upper bounds are equal as known; standard errors and
# covariance matrix entries will be zero for these, and those for
# other parameters should reflect this (so when calculating 
# inverse Hessian, for example, need to select only the 
# portion corresponding to unknown parameters).
#
  SE <- NULL
  cormat <- NULL
  V <- NULL
  Hessian <- NULL
  Obj.thresh <- NULL
  known.pars <- (lb == ub)
  np.unknown <- sum(!known.pars)
  if (is.null(covmat) & stderr) {
   warning("Can't compute standard errors with covmat=NULL")
  } else if (stderr) {
#
# First up: matrix of derivatives of properties wrt parameters. Need 
# to be a bit cunning to work through the list of properties - define a 
# dummy "timescales" list that picks out each property in turn. The use 
# of ts.list below works by creating a list with the right attributes,
# and then replacing each of its values in turn. Since the Ac
# component of timescales is a matrix, it has to be dealt with
# separately.
#
   k <- length(observed)
   tau.deriv <- matrix(nrow=k,ncol=np)
   cur.row <- 0
   bigstep <- min(abs(c(Theta-lolim,hilim-Theta,Theta)))*1e-3
   for (i in 1:length(timescales)) {
    ts.list <- timescales[i]
    if (is.vector(timescales[[i]])) {
     nprops <- length(timescales[[i]])
    } else {
     nprops <- dim(timescales[[i]])[1]
    }    
    for (j in 1:nprops) {
     if (is.vector(timescales[[i]])) {
      ts.list[[1]] <- timescales[[i]][j]
     } else {
      ts.list[[1]] <- matrix((timescales[[i]])[j,],nrow=1)
     }
     cur.row <- cur.row + 1
     tau.deriv[cur.row,] <- num.deriv(f=grad.fn,x0=Theta,xlo=lb,xhi=ub,
                    model=model.name,timescales=ts.list,
                    log.pars=log.pars,log.props=log.props)$gradient
    } 
   }
   W <- wmat %*% tau.deriv
   W <- W[,!known.pars]
#
# Next, calculate the Hessian, either numerically or using an
# analytic and asymptotically equivalent approximation. For 
# numerical evaluation, since the optimisation was done on the 
# transformed parameters, it's probably worth calculating the 
# Hessian wrt these transformed parameters as well, and then 
# back-transforming to the original scale.
#
  if (numeric.hess) {
   phi.lo <- rep(-Inf,np); phi.lo[known.pars] <- 0
   phi.hi <- -phi.lo
   Hessian <- num.hess(f=objfunc,x0=phi,xlo=phi.lo,xhi=phi.hi,
 			observed=observed,w=w,backtrans=TRUE,
 			timescales=timescales,model.name=model.name,
 			lb=lolim,ub=hilim,log.pars=log.pars,
                  log.props=log.props)
   Hess.err <- Hessian$error; Hessian <- Hessian$hessian
   H <- Hessian[!known.pars,!known.pars]
   H <- H * (delta[!known.pars] %*% t(delta[!known.pars]))
   diag.hess <- diag(H); warning.needed <- FALSE
#
# Check that diagonal elements are positive - if not, it's either
# due to inaccuracy in the Hessian calculation or to the fact that
# we're not quite at a minimum. Either way, the options are (a)
# recalculate (expensive and not guaranteed to work) (b) set to
# zero (cheap and cheerful, and will lead to infinite standard
# errors etc.). Since if this occurs, we're probably being a bit
# overambitious, go for (b).
#  
   if (any(!(diag.hess > 0))) {
    warning("Hessian at alleged minimum has negative diagonal element")
    H[!(diag.hess > 0),] <- 0
    H[,!(diag.hess > 0)] <- 0
   } else {
# 
# Otherwise, check that the computed Hessian is positive definite,
# to avoid such embarrassments  as negative objective function
# thresholds ... this is done simply by inflating the diagonal a
# bit if necessary (lacks elegance but it works!)
#
    scale.factor <- 1.0001
    while(isTRUE(class(try(chol(H),silent=TRUE)) == "try-error")) {
     warning.needed <- TRUE
     diag(H) <- scale.factor*diag.hess
     scale.factor <- scale.factor*scale.factor
    }
    if (warning.needed) {
      warning(paste("Diagonal elements of Hessian inflated by",
                      scale.factor,"\nto ensure positive definiteness"))
    }
   }
  } else if (!numeric.hess) {
#
# The analytic approximation to the Hessian is guaranteed +ve
# (semi-)definite by construction, so no particular subtlety 
# needed here
#
   Hessian <- matrix(0,nrow=np,ncol=np)
   H <- 2*t(tau.deriv) %*% wmat %*% tau.deriv
   H <- H[!known.pars,!known.pars]
  }
  rownames(Hessian) <- par.names
  colnames(Hessian) <- par.names
  Hessian[!known.pars,!known.pars] <- H
#
# Now here's the computation of covariance matrix. To get at 
# correlations, use matrix multiplication to generate a matrix with
# products of standard deviations in it, and then elementwise
# division to convert covariance into correlation. Also, do 
# calculations using reduced parameter of unknown paramaters
#
   J <- 4*t(W) %*% covmat %*% W
   Hinv <- try(solve(H),silent=TRUE)
   if (isTRUE(class(Hinv) == "try-error")) {
    warning("Hessian is singular - standard errors will be infinite")
    Hinv <- matrix(Inf,nrow=np.unknown,ncol=np.unknown)
   }
   V <- matrix(0,nrow=np,ncol=np)
   if (is.infinite(Hinv[1,1])) {
    V[!known.pars,!known.pars] <- Hinv
   } else {
    V[!known.pars,!known.pars] <- Hinv %*% J %*% Hinv
   }
   SE <- sqrt(diag(V))
   cor.denom <- SE %*% t(SE)
   cormat <- V / cor.denom
   cormat[known.pars,] <- 1; cormat[,known.pars] <- 1
   rownames(cormat) <- par.names
   colnames(cormat) <- par.names
#
# Approximate thresholds for defining confidence intervals based
# on the objective function itself. Based on a chi-squared 
# approximation to a quadratic form.
#
   M <- Hinv %*% J ; k1 <- sum(diag(M))
   M2 <- M %*% M; k2 <- 2*sum(diag(M2))
   k3 <- 8*sum(diag(M2%*%M))
   a <- abs(k3)/(4*k2)
   b <- 8*(k2^3)/(k3^2)
   c <- k1 - (a*b)
   Obj.thresh <- Obj.func + (c+(a*qchisq(conf.level,b)))/2
   names(Obj.thresh) <- paste(100*conf.level,"%",sep="")
  }
  list(Theta=Theta,Std.err=SE,Corr=cormat,Cov=V,Obj=Obj.func,
	Obj.thresh=Obj.thresh,Gradient=Gradient,Hessian=Hessian,
	fits.table=fits,Bounds=limtable[,2:3],Model=model.name,
	Timescales=timescales,Observed=observed,
	w=w,log.pars=rep(FALSE,np)|log.pars,
        log.props=rep(FALSE,length(observed))|log.props)
}
#########################################################################
# 	Optimisation routine, with options
#	
#	Arguments:
#	theta		Starting value for parameter vector
#	observed	Vector of observed statistics
#	w		Either a vector of weights to associate with 
#			observed statistics, or a weighting matrix
#	method	Method to use for numerical minimisation.
#			If this is a character vector, the routine
#			will try each method in turn.
#	ntries	Number of different starting points to try
#	print.level	Controls amount of output. Values:
#			0	Work silently
#			1	Print a message each time a new fit
#				is started
#			2	Print estimates and objective function
#				values from each fit.
#	lb		Lower bounds on parameters. Defaults to 
#			`almost zero', to avoid problems on the 
#			boundary itself when you divide by a parameter
#	ub		Upper bounds on parameters
#	log.pars	Logical scalar or vector, indicating which
#			(if any) of the parameters in theta have been 
#			log-transformed prior to calling this function.
#			This is included primarily for use with
#			auto.fit.
#     log.props   Logical scalar or vector, indicating which (if
#                 any) of the components of observed are estimates
#                 of the corresponding log properties.
#	gradients	If TRUE, numerical gradients will be calculated
#			at the final estimates. Otherwise, gradients
#			will be set to NA.
#	plot		If TRUE, plots are produced to show the best
#			objective function values found.
#########################################################################
moment.fit <- function(model.name,theta,timescales,observed,
                       w=rep(1,length(theta)),method="Nelder-Mead",
                       ntries=1,print.level=1,lb=1e-6,ub=Inf,
                       log.pars=FALSE,log.props=FALSE,gradients=TRUE,
                       plot=TRUE) {
 optim.methods <- c("Nelder-Mead","BFGS","CG","L-BFGS-B","SANN")
 np <- length(theta); nmethods <- length(method)
 equal.bounds <- lb == ub
 for (p in 1:np) {
  if (lb[p] >= theta[p] & !equal.bounds[p]) {
   stop(paste("\nAll starting values must be strictly greater",
              "than lower bounds -","\nfor parameter",p,
              "the starting value is",theta[p],
              "\nbut lower bound is",lb[p]))
  } else if (ub[p] <= theta[p] & !equal.bounds[p]) {
   stop(paste("\nAll starting values must be strictly less",
              "than upper bounds -","\nfor parameter",p,
              "the starting value is",theta[p],
              "\nbut upper bound is",ub[p]))
  }
 } 
 z <- data.frame(matrix(nrow=ntries*nmethods,ncol=2*np+3))
 names(z) <- c(paste(rep(c("Theta","Grad"),np),rep(1:np,each=2),sep=""),
					"Obj","Method","Converge")
#
#     Define 'typical' magnitudes of parameters, and target value
#     of objective function
#
 par.size <- abs(par.trans(theta,lb,ub))
 ini.val <- objfunc(theta,backtrans=FALSE,lb=lb,ub=ub,observed=observed,
                    w=w,timescales=timescales,model.name=model.name,
                    log.pars=log.pars,log.props=log.props)
 if (!is.null(names(theta))) {
   names(z)[seq(1,2*np,2)] <- names(theta)
 }
 bigstep <- 2*min(0.5,max(abs(log(abs(theta)))))
#
#	Cycle through the requested methods, transforming 
#	constrained parameters to the whole real line first.
#
 for (meth.idx in 1:length(method)) {
  meth <- method[meth.idx]
  useLBFGSB <- (meth == "L-BFGS-B")
  for (i in 1:ntries) {
   if (print.level > 0) {
    cat(paste("Minimisation method ",meth,
		", starting point ",i,"\n",sep=""))
   }
   zrow <- ntries*(meth.idx-1) + i
   fixed <- (i == 1)
   startval <- par.trans(theta,lb,ub,"forward",fixed,useLBFGSB)
   if (meth %in% optim.methods) {
    if (useLBFGSB) {
     thetahat <- try(optim(startval,objfunc,method=meth,
                           lower=lb,upper=ub,
                           backtrans=FALSE,observed=observed,w=w,
                           timescales=timescales,model.name=model.name,
                           log.pars=log.pars,log.props=log.props),
                           silent=TRUE)
    } else {
     thetahat <- try(optim(startval,objfunc,method=meth,
                           lb=lb,ub=ub,backtrans=TRUE,
                           observed=observed,w=w,
                           timescales=timescales,model.name=model.name,
                           log.pars=log.pars,log.props=log.props),
                           silent=TRUE)
    }
    ierr <- isTRUE(class(thetahat) == "try-error")
    if (!ierr) {
     phi <- thetahat$par
     estimate <- par.trans(phi,lb,ub,"backward",TRUE,useLBFGSB)
     fmin <- thetahat$value
     conv <- thetahat$convergence == 0
    }
   } else if (meth %in% c("nlm","NLM")) {
#
#	For nlm method, as we're working on log scale we need to 
#	control the size of step to be taken. 
#
    thetahat <- try(nlm(objfunc,startval,stepmax=bigstep,
			typsize=par.size,fscale=ini.val/10,
                        lb=lb,ub=ub,backtrans=TRUE,observed=observed,
                        w=w,timescales=timescales,
                        model.name=model.name,
			log.pars=log.pars,log.props=log.props),silent=TRUE)
    ierr <- isTRUE(class(thetahat) == "try-error")
    if (!ierr) {
     phi <- thetahat$estimate
     estimate <- par.trans(phi,lb,ub,"backward",TRUE,useLBFGSB)
     fmin <- thetahat$minimum
     conv <- (thetahat$code <= 2)
    }
   }
   if (ierr) {
    estimate <- rep(NA,np)
    phi <- estimate
    fmin <- NA
    grad <- estimate
    conv <- FALSE    
   } else {
#
#	Calculate gradients if requested. Accuracy isn't critical,
#	so limit number of iterations in num.deriv(). NB however
#	to avoid stability problems at boundaries, carry out 
#	differentiation wrt *transformed* parameters and then
#	back-transform the calculated gradients
#
    grad <- rep(NA,np)
    if (gradients) {
     grad <- num.deriv(f=objfunc,x0=phi,maxit=3,observed=observed,
 			w=w,backtrans=TRUE,timescales=timescales,
  			model.name=model.name,lb=lb,ub=ub,
  			log.pars=log.pars,log.props=log.props)
     grad <- grad$gradient * grad.trans(phi,lb,ub,useLBFGSB)
    }
   }
   if (print.level > 1) {
    cat("Estimate: "); cat(round(estimate,3))
    cat(paste("\nObjective function value:",round(fmin,4),"\n"))
   }
   z[zrow,(2*(1:np)-1)] <- estimate
   z[zrow,2*(1:np)] <- grad
   z$Obj[zrow] <- fmin
   z$Method[zrow] <- meth
   z$Converge[zrow] <- conv
  }
 }
 if (plot) plot.mmfit(z)
 z
}
#########################################################################
#	Transformation of parameter vector. This is used by
#	moment.fit to incorporate limits into estimation procedure, 
#	and by objfunc to recover original parametrisation from 
#	transformed values. It is not designed for user interaction!
#
#	Arguments:
#
#	theta		Parameter vector
#	lolim		Vector of lower bounds on elements of theta
#	hilim		Vector of upper limits
#	direction	Controls direction of transformation -
#			"forward" to compute transformed 
#			values, "backward" to recover originals
#	fixed		Take theta as fixed, or generate a random
#			perturbation about it (only used for
#			direction="forward")
#	useLBFGSB	Indicates whether we're using the 
#			LBFGSB optimisation method in 'optim' -
#			if so, let the optimisation method deal
#			with the bounds directly.
#########################################################################
par.trans <- function(theta,lolim=-Inf,hilim=Inf,
		direction="forward",fixed=TRUE,useLBFGSB=FALSE) {
 np <- length(theta)
 if (length(lolim) == 1) lb <- rep(lolim,np) else lb <- lolim
 if (length(hilim) == 1) ub <- rep(hilim,np) else ub <- hilim
 equal.bounds <- lb == ub
 if (direction == "forward" ) {
  f <- theta
  if (any(f < lb)) stop("Parameter is below lower limit")
  if (any(f > ub)) stop("Parameter is above upper limit")
#
#	When one limit is finite, use exponential sampling to
#	generate random perturbations
#
  want <- is.finite(lb) & is.infinite(ub)
  f[want] <- theta[want]-lb[want]
  if (!fixed) f[want] <- rexp(sum(want),rate=1/f[want])
  if (!useLBFGSB) {
   f[want] <- log(f[want]) 
  } else { 
   f[want] <- f[want] + lb[want]
  }
  want <- is.finite(ub) & is.infinite(lb)
  f[want] <- ub[want] - theta[want]
  if (!fixed) f[want] <- rexp(sum(want),rate=1/f[want])
  if (!useLBFGSB) {
   f[want] <- log(f[want])
  } else {
   f[want] <- ub[want] - f[want]
  }
#
#	For finite upper and lower limits, use uniform sampling.
#	Need special treatment when bounds are equal (i.e.
#	parameter is fixed) to avoid dividing by zero.
#
  want <- is.finite(ub) & is.finite(lb)
  if (fixed) {
   f[want] <- theta[want]
  } else {
    f[want] <- runif(sum(want),lb[want],ub[want])
  }
  if (!useLBFGSB) {
   want2 <- want & !equal.bounds
   f[want2] <- log(log( (ub[want2]-lb[want2]) / 
			                (f[want2]-lb[want2]) ))
   f[want & equal.bounds] <- 0
  }
 } else if (direction == "backward") {
  f <- theta
  if (!useLBFGSB) {
   want <- is.finite(lb) & is.infinite(ub)
   f[want] <- lb[want] + exp(theta[want])
   want <- is.finite(ub) & is.infinite(lb)
   f[want] <- ub[want] - exp(theta[want])
   want <- is.finite(ub) & is.finite(lb) & !equal.bounds
   f[want] <- lb[want] + (ub[want]-lb[want])*exp(-exp(theta[want]))
   want <- is.finite(ub) & is.finite(lb) & equal.bounds
   f[want] <- lb[want]
#
# Correct for any rounding that may have caused f to fall outside
# the requested interval
#
   f[f < lb] <- lb[f < lb]; f[f > ub] <- ub[f > ub]
  }
 }
 f
}
#########################################################################
#	Derivatives of parameter vector transformation: specifically
#     the quantities d(phi)/d(theta), where theta are the original 
#	parameters and phi are transformed (i.e. 
#     phi = par.trans(theta,...,direction="forward")). 
#
#	Arguments:
#
#	phi		The phi-value at which to evaluate the gradient
#	lolim		Vector of lower bounds on elements of theta
#	hilim		Vector of upper limits
#	useLBFGSB	Indicates whether we're using the LBFGSB
#			optimisation method in 'optim' - if so, the
#			optimisation method will have dealt with
#			the bounds directly.
#
#########################################################################
grad.trans <- function(phi,lolim=-Inf,hilim=Inf,useLBFGSB=FALSE) {
 np <- length(phi)
 theta <- par.trans(phi,lolim,hilim,"backward",TRUE,useLBFGSB)
 if (length(lolim) == 1) lb <- rep(lolim,np) else lb <- lolim
 if (length(hilim) == 1) ub <- rep(hilim,np) else ub <- hilim
 if (any(theta < lb)) stop("Parameter is below lower limit")
 if (any(theta > ub)) stop("Parameter is above upper limit")
 equal.bounds <- lb == ub
#
#	For untransformed parameters, gradients are all 1. For
#	equal bounds, gradients are infinite (any value of phi
#	gives the same theta, so d(theta)/d(phi) = 0)
#
 f <- rep(1,np); f[equal.bounds] <- Inf

 if (!useLBFGSB) {
#
#	Finite lower bound
#
  want <- is.finite(lb) & is.infinite(ub)
  f[want] <- exp(-phi[want])
#
#	Finite upper bound
#
  want <- is.finite(ub) & is.infinite(lb)
  f[want] <- -exp(-phi[want])
#
#	Finite upper and lower bounds
#	
  want <- is.finite(ub) & is.finite(lb)
  f[want] <- -exp(-phi[want]) / (theta[want] - lb[want])
 }
 f
}
######################################################################
#     Plot the table of results from a fitting exercise.   
#	Arguments:
#
#       fit.tab     Table of fitting results
#       pltfrac     If we've got enough `converged' points in fit.tab,
#                   plot the best 100*pltfrac% of these,  then add any 
#	            `non-converged' points that will fit. Otherwise plot the
#	            best 100*pltfrac% of all fits.
#       xlimits     Optional table of x limits for plots (to ensure that
#                   all points appear)
#       nonconvergedIf TRUE, axes will be set up to accommodate
#                   fits that are not deemed to have converged.
#       plot.grad   Do we want to add lines to each point on the plot,
#                   indicating the calculated gradients?
#       linlen      Length of gradient lines on the plot, relative to
#                   a default setting that is intended to look reasonable
#                   but can't be guaranteed for all graphics layouts
######################################################################
plot.mmfit <- function(fit.tab,pltfrac=0.5,xlimits=NULL,
                       nonconverged=FALSE,plot.grad=FALSE,linlen=1) {
  z <- fit.tab; np <- (dim(z)[2] - 3)/2; nobs <- dim(z)[1]
  old.pty <- par("pty")
  if (plot.grad) par(pty="s")
  if (!is.null(colnames(z))) {
#
#   parse() sometimes complains for no apparent reason! In this
#   case set default plot titles
#
    plot.titles <- try(parse(text=colnames(z)[seq(1,2*np,2)]),
                       silent=TRUE)
    if (isTRUE(class(plot.titles) == "try-error")) {
      plot.titles <- paste("Parameter",1:np)
    }
  } else {
    plot.titles <- paste("Parameter",1:np)
  }
  want <- z$Converge
  if (sum(want) > 0) {
    if (nonconverged) {
      thresh <- quantile(z$Obj,probs=pltfrac,na.rm=TRUE)
    } else {
      thresh <- quantile(z$Obj[want],probs=pltfrac,na.rm=TRUE)
    }
    want <- want & z$Obj <= thresh
    pch1 <- 16
  } else {
    pch1 <- 1
    thresh <- quantile(z$Obj,probs=pltfrac,na.rm=TRUE)
    want <- z$Obj <= thresh
 }
  pch2 <- 1
  for (p in 1:np) {
   colno <- 2*p - 1
   if (is.null(xlimits) || any(is.infinite(as.numeric(xlimits[p,])))) {
     if (nonconverged) {
       xlim <- range(z[z$Obj < thresh,colno])
     } else {
       xlim <- range(z[want,colno])
     }
   } else {
     xlim <- as.numeric(xlimits[p,])
   }
   ylim <- c(min(z$Obj,na.rm=TRUE),thresh)
   plot(z[want,colno],z$Obj[want],xlab="Estimate",
        xlim=xlim,ylim=ylim,ylab="Objective function",
        pch=pch1,main=plot.titles[p])
   points(z[!z$Converge,colno],z$Obj[!z$Converge],pch=pch2)
   if (plot.grad) {
    xlen <- 0.2*linlen*(xlim[2]-xlim[1])/sqrt(nobs)
    ylen <- 0.2*linlen*(ylim[2]-ylim[1])/sqrt(nobs)
    rot <- atan(z[,colno+1])
    dx <- xlen*cos(rot); dy <- ylen*sin(rot)
    for (i in 1:nobs) {
     lines(c(z[i,colno]-dx[i],z[i,colno]+dx[i]),
           c(z$Obj[i]-dy[i],z$Obj[i]+dy[i]))
    }
   }
 }
 par(pty=old.pty) 
}
######################################################################
#       `Profile' objective function plots. Arguments:
#
#	fit		The output of a call to auto.fit(), with a 
#			non-null value of Obj.thresh (see documentation 
#			for auto.fit() for details)
#	params		Vector of parameters for which profiles are 
#			required.
#	grids		An optional list of vectors containing, for
#			each parameter, a grid of values at which
#			to calculate the objective function. If
#			NULL (the default), a regular grid between
#			loplim and upplim is used. NOTE that even
#			if some of the parameters in fits are on a 
#			log scale, all elements of grids should be 
#			on the *original* scale. Ditto loplim and
#			upplim
#	loplim  	Vector of lower plotting limits for each 
#			parameter. Defaults to estimate - 3 standard
#			errors.
#	upplim		Vector of upper plotting limits. Defaults to 
#			estimate + 3 standard errors
#	lower   	Vector of lower bounds on each parameter, to 
#			constrain optimisation
#	upper   	Vector of upper bounds, similarly 
#	npoints 	Number of intermediate points at which to 
#			calculate the profile objective function
#			for each parameter
#	conf.level	A vector of probabilities from which to
#			compute objective function thresholds 
#			corresponding to approximate confidence intervals. 
#			The default value of c(0.95,0.99) returns 
#			thresholds corresponding to approximate 95% 
#			and 99% intervals.
#	plot.it		Logical determining whether or not to plot
#			the results as we go.
#	hlines  	Vector selecting which of the thresholds 
#			corresponding to conf.level to plot as horizontal 
#			lines. Defaults to all of them.
#	lintyps 	Vector of line types for the various thresholds.
#			Defaults to 1+(1:length(hlines))
#	print.level 	Controls verbosity of output
#	ninit		Controls behaviour of auto.fit } See auto.fit
#	nsubseq  	   "        "      "     "     } header for
#	nsearches          "        "      "     "     } details
######################################################################
obj.profile <- function(fit,params,grids=NULL,loplim=NULL,upplim=NULL,
			lower=rep(1e-6,length(fit$Theta)),
			upper=rep(Inf,length(fit$Theta)),npoints=10,
			conf.level=c(0.95,0.99),plot.it=TRUE,
			hlines=1:length(conf.level),
			lintyps=1+(1:length(hlines)),print.level=1,
			ninit=2,nsubseq=10,nsearches=1) {
 if (is.null(fit$Cov)) {
  stop("'fit' object does not contain covariance matrix")
 }
 if (length(lintyps) != length(hlines)) {
  warning("Number of line types differs from number of lines")
  lintyps <- 1+(1:length(hlines))
 }
 if (any(upper < lower)) stop("upper bound < lower bound")
#
# Check that plotting limits are compatible with constraints. At
# this point, put constraints on log scale where necessary
#
 logxfrm <- fit$log.pars
 lp <- lower; lp[logxfrm] <- log(lp[logxfrm])
 up <- upper; up[logxfrm] <- log(up[logxfrm])
 lp <- lp[params]; up <- up[params]
 par.adj <- (up-lp)*1e-4; par.adj[is.infinite(par.adj)] <- 1e-4
 lpdef <- fit$Theta-(3*fit$Std.err); lpdef <- lpdef[params]
 updef <- fit$Theta+(3*fit$Std.err); updef <- updef[params]
 if (is.null(grids)) {
  if (any(upplim <= loplim)) stop("upper plotting limits must exceed lower ones")
  if (is.null(loplim)) {
   loplim <- lpdef
   if (any(loplim < lp)) {
    warning(paste("Resetting lower plotting limits where these are <",
                  "lower bounds on parameters\n"))
    loplim[loplim < lp] <- (lp+par.adj)[loplim < lp]
   }
  } else {
   loplim[logxfrm[params]] <- log(loplim[logxfrm[params]])
   if (any(loplim < lp)) {
    warning("Some lower plotting limits are < lower bounds on parameters\n")
   }
  }
  if (is.null(upplim)) {
   upplim <- updef
   if (any(upplim > up)) {
    warning(paste("Resetting upper plotting limits where these are >",
                  "upper bounds on parameters\n"))
    upplim[upplim > up] <- (up-par.adj)[upplim > up]
   } 
  } else {
   upplim[logxfrm[params]] <- log(upplim[logxfrm[params]])
   if (any(upplim > up)) {
    warning("Some upper plotting limits are > upper bounds on parameters\n")
   }
  }
 } else {
  if (!is.list(grids)) stop("grids must be a list")
  for (ip in 1:length(params)) {
   if (logxfrm[[ip]]) grids[[ip]] <- log(grids[[ip]])
   if (any(grids[[ip]] < lp[ip])) {
    warning("Some grid values are < lower bounds on parameters\n")
   } 
   if (any(grids[[ip]] > up[ip])) {
    warning("Some grid values are > upper bounds on parameters\n")
   } 
  }
 }
#
# Define plot titles. parse() sometimes complains for no apparent
# reason - in this case set default plot titles
#
 if (!is.null(names(fit$Theta))) {
   plot.titles <- try(parse(text=names(fit$Theta)),silent=TRUE)
   if (isTRUE(class(plot.titles) == "try-error")) {
     plot.titles <- paste("Parameter",1:length(fit$Theta))
   }
 } else {
  plot.titles <- paste("Parameter",1:length(fit$Theta))
 }
#
# Suppress printing of warning messages
#
 warnings.value <- getOption("warn"); options(warn=-1)
#
# Calculations for objective function thresholds - treat as
# known any values with a zero variance
#
 V <- fit$Cov; H <- fit$Hessian
 known.pars <- fit$Std.err == 0
 if (any(known.pars[params])) {
  stop("Profile requested for a parameter that has been held fixed\n")
 }
 k1 <- rep(Inf,length(fit$Theta))
 k1[!known.pars] <- diag(V)[!known.pars] / 
                     diag(solve(H[!known.pars,!known.pars]))
 k1 <- k1[params]
 thresh.names <- paste(100*conf.level,"%",sep="")
#
# Loop over parameters
#
 z <- list()
 for (ip in 1:length(params)) {
  p <- params[ip]
  txtlab <- names(fit$Theta)[p]
  if (print.level > 0) {
   cat(paste("Calculating profile objective function for",
             					txtlab,"...\n"))
  }
  if (is.null(grids)) {
   xint <- (upplim[ip]-loplim[ip])/npoints
   x <- sort(c(fit$Theta[p],seq(loplim[ip],upplim[ip],xint)))
  } else {
   x <- sort(c(grids[[ip]],fit$Theta[p]))
  }
  x <- unique(x); y <- rep(0,length(x))
  minpos <- match(fit$Theta[p],x)
  y[minpos] <- fit$Obj
  theta.grid <- matrix(nrow=length(x),ncol=length(fit$Theta))
  theta.grid[minpos,] <- fit$Theta
  if (logxfrm[p]) x <- exp(x)
#
# Start from the optimum and work downwards, at each stage using the 
# previous estimates as starting values (but avoiding boundaries,
# if the estimate was exactly on a boundary). Shouldn't need too many
# trials since the starting value should be good.
#
  for (i in seq(minpos-1,1,-1)) {
   xout <- x[i]; if (logxfrm[p]) xout <- log(xout)
   if (print.level > 0) cat(paste(txtlab,"=",signif(xout,5),"\t"))
   cur.theta <- theta.grid[i+1,]
   unusable <- is.na(cur.theta) | is.infinite(cur.theta)
   cur.theta[unusable] <- fit$Theta[unusable]
   cur.theta[logxfrm] <- exp(cur.theta[logxfrm])
   lb <- lower; ub <- upper; tol <- 1e-7
   lbclose <- (cur.theta - lb < tol)
   cur.theta[lbclose] <- lb[lbclose] + (10*tol*(ub[lbclose]-lb[lbclose]))
   ubclose <- (ub - cur.theta < tol)
   cur.theta[ubclose] <- ub[ubclose] - (10*tol*(ub[ubclose]-lb[ubclose]))
   cur.theta[p] <- x[i]
   lb[p] <- x[i]; ub[p] <- x[i]
   cur.fit <- auto.fit(model.name=fit$Model,theta=cur.theta,
                       timescales=fit$Timescales,observed=fit$Observed,
                       w=fit$w,print.level=print.level-1,
                       lb=lb,ub=ub,ninit=ninit,nsubseq=nsubseq,
                       nsearches=nsearches,plot.progress=FALSE,
                       log.pars=logxfrm,log.props=fit$log.props)
   y[i] <- cur.fit$Obj
   if (print.level > 0) cat(paste("Objective function =",signif(y[i],5),"\n"))
   theta.grid[i,] <- cur.fit$Theta
  }
#
# Now go back to the optimum and work upwards
#
  for (i in seq(minpos+1,length(x),1)) {
   xout <- x[i]; if (logxfrm[p]) xout <- log(xout)
   if (print.level > 0) cat(paste(txtlab,"=",signif(xout,5),"\t"))
   cur.theta <- theta.grid[i-1,]
   unusable <- is.na(cur.theta) | is.infinite(cur.theta)
   cur.theta[unusable] <- fit$Theta[unusable]
   cur.theta[logxfrm] <- exp(cur.theta[logxfrm])
   lb <- lower; ub <- upper
   lbclose <- (cur.theta - lb < tol)
   cur.theta[lbclose] <- lb[lbclose] + (10*tol*(ub[lbclose]-lb[lbclose]))
   ubclose <- (ub - cur.theta < tol)
   cur.theta[ubclose] <- ub[ubclose] - (10*tol*(ub[ubclose]-lb[ubclose]))
   cur.theta[p] <- x[i]
   lb[p] <- x[i]; ub[p] <- x[i]
   cur.fit <- auto.fit(model.name=fit$Model,theta=cur.theta,
                       timescales=fit$Timescales,observed=fit$Observed,
                       w=fit$w,print.level=print.level-1,stderr=FALSE,
                       lb=lb,ub=ub,ninit=ninit,nsubseq=nsubseq,
                       nsearches=nsearches,plot.progress=FALSE,
                       log.pars=logxfrm,log.props=fit$log.props)
   y[i] <- cur.fit$Obj
   if (print.level > 0) cat(paste("Objective function =",signif(y[i],5),"\n"))
   theta.grid[i,] <- cur.fit$Theta
  }
#
# Calculate objective function thresholds for this parameter
#
  thresholds <- fit$Obj + (k1[ip]*qchisq(conf.level,1)/2)
  thresh.tab <- matrix(rep(thresholds,length(x)),nrow=length(x),byrow=TRUE)
#
# If requested, produce plot for this parameter
#
  if (plot.it) {
   ymax <- max(y,thresholds[hlines],na.rm=TRUE) 
   ymin <- min(y,fit$Obj,na.rm=TRUE) 
   plot(theta.grid[,p],y,ylim=c(ymin,ymax),xlab="Parameter value",
        ylab="Objective function",type="l",lwd=2)
   title(main="Profile objective function")
   title(main=plot.titles[p],line=1)
   for (t in hlines) abline(thresholds[t],0,lwd=2,lty=lintyps[t])
   legend(min(theta.grid[,p]),ymax,thresh.names[hlines],
          lty=lintyps,bg="white",lwd=2)
  }
#
# And store the results table in the corresponding element of z
#
  theta.grid <- data.frame(cbind(theta.grid,y,thresh.tab))
  names(theta.grid) <- c(names(fit$Theta),"Obj.func",thresh.names)
  z[[ip]] <- theta.grid
 }
 names(z) <- names(fit$Theta)[params]
 options(warn=warnings.value)
 z
}

######################################################################
#     Numerical differentiation (because there aren't any 
#	user-friendly routines for this in R). Arguments:
#
#	f	The function to be differentiated
#	x0	The point at which to evaluate the gradient
#	lb	Vector of lower bounds on elements of x (either
#		a scalar, or a vector the same length as x0)
#	ub	Vector of upper bounds
#	steps	Vector of largest step sizes to consider when
#		evaluating the gradient. Defaults to the larger
#		of abs(x0)/10^4 and 0.001.
#	shrink	Amount by which to shrink the stepsize at each
#		iteration, to obtain an improved estimate
#       maxit   Maximum number of iterations
#       relerr  Relative error required
#	...	Any other arguments to f
#
#     This uses Ridders' method - see Press et al. (1992), Section
#	5.7 for example. The implementation here is basically that
#	given by Press et al., translated from FORTRAN to R and
#     adapted to handle boundaries (in which case a right- or 
#	left-hand derivative is evaluated as appropriate) and 
#	vector arguments. 1-sided derivatives are calculated by 
#     reflecting the function about the boundary, which is
#	equivalent to using function evaluations on one side of the
#	boundary only - and correspondingly less accurate. For
#	1-sided derivatives it may therefore be necessary to 
#	decrease the default value of steps to obtain the required 
#	accuracy.
#
#	The value of the function is a list containing two elements,
#	as follows:
#
#	gradient	vector of length length(x0), containing
#			the calculated derivatives
#	error		Estimate of the errors in the computed
#			gradients
#	
######################################################################
num.deriv <- function(f,x0,xlo=-Inf,xhi=Inf,steps=NULL,shrink=1.4,
					maxit=10,relerr=1e-6,...) {
 np <- length(x0); d <- numeric(np)
 nb <- length(xlo)
 if (nb > 1 & nb != np) stop("xlo and x0 must be the same length")
 nb <- length(xhi)
 if (nb > 1 & nb != np) stop("xhi and x0 must be the same length")
 if (is.null(steps)) {
  h <- pmax(abs(x0)*1e-4,1e-3) 
 } else {
  h <- steps
 }
 right <- (x0 == xlo); left <- (x0 == xhi)
 f0 <- f(x0,...)
 shr2 <- shrink^2; err <- rep(Inf,np)
 delta <- .Machine$double.eps
#
# Set initial step sizes for each parameter. 
#
 hh <- pmin(h,(x0-xlo-delta),(xhi-delta-x0))
 hh[left] <- pmin(h[left],(xhi-xlo-delta)[left])
 hh[right] <- pmin(h[right],(xhi-xlo-delta)[right])
 tight <- right & left; d[tight] <- NA
 wanted.els <- (1:np)[!tight]
 for (p in wanted.els) {
  a <- matrix(nrow=maxit,ncol=maxit)
  i <- 1; cnvrge <- FALSE
  while (i <= maxit & !cnvrge) {
   xx <- x0
   if (!right[p]) {
    xlo <- x0[p] - hh[p]; xx[p] <- xlo; f1 <- f(xx,...)
#
#	For left derivatives, extrapolate symmetrically beyond the
#	upper limit
#
    if (left[p]) f2 <- (2*f0) - f1
   }
   if (!left[p]) {
    xhi <- x0[p] + hh[p]; xx[p] <- xhi; f2 <- f(xx,...)
#
#	And for right derivatives, extrapolate beyond the lower
#	limit
#
    if (right[p]) f1 <- (2*f0) - f2
   }
   a[1,i] <- (f2 - f1) / (2*hh[p])
#
#	This is the j-loop in Numerical Recipes. Although looping
#	is normally to be avoided in R, experimentation suggests
#	that in this case it's actally quicker than the corresponding
#	vector-based calculation - partly do do with the fact that 
#	in that case you have do do things like identify the 
#	location of the smallest value in the column etc., I guess
#
   if (i > 1) {
    fac <- shr2
    for (j in 2:i) {
     a[j,i] <- (a[j-1,i]*fac-a[j-1,i-1]) / (fac-1)
     fac <- shr2*fac
     errt <- max(abs(a[j,i]-a[j-1,i]),abs(a[j,i]-a[j-1,i-1]))
     if (is.na(errt)) errt <- Inf
     if (errt < err[p]) { 
      err[p] <- errt
      d[p] <- a[j,i]
     }
    }
     cnvrge <- (abs(a[i,i]-a[i-1,i-1]) > 2*err[p])
     if (is.na(cnvrge)) cnvrge <- FALSE
     if (abs(d[p]) > relerr) {
      cnvrge <- cnvrge & (abs(err[p]) < abs(relerr * d[p]))
     }
    }
   hh[p] <- hh[p] / shrink; i <- i+1
  }
 }
 list(gradient=d,error=err)
}
######################################################################
#       Numerical evaluation of a Hessian. Arguments:
#
#	f	The function to be differentiated
#	x0	The point at which to evaluate the Hessian
#	xlo	Vector of lower bounds on elements of x (either
#		a scalar, or a vector the same length as x0)
#	xhi	Vector of upper bounds
#	steps	Vector of largest step sizes to consider for 
#		finite-difference approximations. Default is
#               approximately the square root of than that in 
#		num.deriv, because the denominator in each 
#		approximation is steps^2)
#	shrink	Amount by which to shrink the stepsize at each
#		iteration, to obtain an improved estimate
#       maxit   Maximum number of iterations
#       relerr  Relative error required
#	...	Any other arguments to f
#
#       This uses polynomial extrapolation to the limit, in exactly
#       the same way as num.deriv(). The result is regarded is the
#       limit of a sequence of finite-difference approximations, 
#       evaluated at a sequence of points e1, e2, e3, ... 
#       which represent progressively smaller multiples of the 
#       initial value of steps. The finite difference approximation 
#	actually used is, for off-diagonal elements,
#
#	{f(x1+h1,x2+h2)+f(x1-h1,x2-h2)-[f(x1+h1,x2-h2)+f(x1-h1,x2+h2)]}
#       _______________________________________________________________
#				4*h1*h2
#
#	or (f1 + f2 - f3 - f4) / (4*h1*h2), where h1 and h2 are 
#	multiples of the e's. For diagonal elements the approximation
#	is
#
#	{f(x+h) - f(x) - f(x) + f(x-h)} / h
#
#	- NB the use of h rather than 4h is *correct* here - repeated
#       application of the first-order finite difference calculation
#       yields a step size of 2h, which has been halved in this case.
#
#	The value of the function is a list containing two elements,
#	as follows:
#
#	hessian	matrix containing the calculated Hessian. For
#                 all variables where the lower and upper bounds 
#			xlo and xhi are equal, a value of zero is returned. 
#	error		Estimate of the errors in the computed
#			hessian
#	
######################################################################
num.hess <- function(f,x0,xlo=-Inf,xhi=Inf,steps=NULL,shrink=1.4,
					maxit=10,relerr=1e-6,...) {
 np <- length(x0); d <- matrix(rep(0,np*np),nrow=np,ncol=np)
 nb <- length(xlo)
 if (nb > 1 & nb != np) stop("xlo and x0 must be the same length")
 nb <- length(xhi)
 if (nb > 1 & nb != np) stop("xhi and x0 must be the same length")
 if (is.null(steps)) {
  h <- pmax(abs(x0)*1e-2,1e-2) 
 } else {
  h <- steps
 }
 right <- (x0 == xlo); left <- (x0 == xhi)
 tight <- right & left
 if (any((right | left)[!tight])) {
   stop("I don't know how to evaluate Hessian on a boundary")
 }
 shr2 <- shrink^2; err <- matrix(rep(Inf,np*np),nrow=np)
 delta <- .Machine$double.eps
 oppsign <- c(1,-1)
 f0 <- f(x0,...)
 for (p1 in (1:np)[!tight]) {
  for (p2 in (1:np)[!tight]) {
#
# Reset initial step sizes for each parameter. No need to set 
# tight values to NA here, because d contains NA on initialisation
#
   hh <- pmin(h,(x0-xlo-delta),(xhi-delta-x0))
   a <- matrix(nrow=maxit,ncol=maxit); els <- c(p1,p2)
   i <- 1; cnvrge <- FALSE
   while (i <= maxit & !cnvrge) {
    xx <- x0
#
# Different finite difference approximations required on and off
# the diagonal
#
    if (p1 == p2) {
     xnew <- x0[p1] - hh[p1]; xx[p1] <- xnew; f1 <- f(xx,...)
     xnew <- x0[p1] + hh[p1]; xx[p1] <- xnew; f2 <- f(xx,...)
     a[1,i] <- ((f1 - f0) - (f0 - f2)) / (hh[p1]*hh[p1])
    } else {
     xnew <- x0[els] + hh[els]; xx[els] <- xnew; f1 <- f(xx,...)
     xnew <- x0[els] - hh[els]; xx[els] <- xnew; f2 <- f(xx,...)
     xnew <- x0[els] + (hh[els]*oppsign); xx[els] <- xnew; f3 <- f(xx,...)
     xnew <- x0[els] - (hh[els]*oppsign); xx[els] <- xnew; f4 <- f(xx,...)
     a[1,i] <- (f1 + f2 - f3 - f4) / (4*hh[p1]*hh[p2])
   }
#
#	This is the j-loop in Numerical Recipes. Although looping
#	is normally to be avoided in R, experimentation suggests
#	that in this case it's actally quicker than the corresponding
#	vector-based calculation - partly do do with the fact that 
#	in that case you have do do things like identify the 
#	location of the smallest value in the column etc., I guess
#
    if (i > 1) {
     fac <- shr2 
     for (j in 2:i) {
      a[j,i] <- (a[j-1,i]*fac-a[j-1,i-1]) / (fac-1)
      fac <- shr2*fac
      errt <- max(abs(a[j,i]-a[j-1,i]),abs(a[j,i]-a[j-1,i-1]))
      if (is.na(errt)) errt <- Inf
      if (errt < err[p1,p2]) { 
       err[p1,p2] <- errt
       d[p1,p2] <- a[j,i]
      }
     }
     cnvrge <- (abs(a[i,i]-a[i-1,i-1]) > 2*err[p1,p2]) 
     if (is.na(cnvrge)) cnvrge <- FALSE
     if (abs(d[p1,p2]) > relerr) {
      cnvrge <- cnvrge & (abs(err[p1,p2]) < abs(relerr * d[p1,p2]))
     }
    }
    hh[els] <- hh[els] / shrink; i <- i+1
   }
  }
 }
 list(hessian=(d+t(d))/2,error=err)
}
